from ._LoyaltyApplierSubType import LoyaltyApplierSubType
from ._LoyaltyButtonEnum import LoyaltyButtonEnum
from ._LoyaltyDiscountData import LoyaltyDiscountData
from ._LoyaltyExceptions import LoyaltyStopPaymentAction



import base64
import logging
import zlib

from abc import ABCMeta
from boto3 import client as BotoClient

from messageprocessorutil import UuidMessageProcessor


logger = logging.getLogger("TopicClient")


class ServerEventProcessor(UuidMessageProcessor):
    __metaclass__ = ABCMeta

    def __init__(
        self,
        event_name,                 # type: str
        store_code,                 # type: str
        context,                    # type: str
        sns_client,                 # type: BotoClient
        sqs_client,                 # type: BotoClient
        topic_arn,                  # type: str
        server_queue_url,           # type: str
        message_maximum_length,     # type: int
    ):
        # type: (...) -> None

        super(ServerEventProcessor, self).__init__(event_name)
        self.store_code = store_code
        self.context = context
        self.sns_client = sns_client
        self.sqs_client = sqs_client
        self.topic_arn = topic_arn
        self.server_queue_url = server_queue_url
        self.message_maximum_length = message_maximum_length

    def parse_data(self, data, event_data):
        return data

    def call_business(self, obj):
        return obj

    def format_response(self, message_bus, message, event, data, result):
        subject = event.subject
        data_length = len(data.encode('utf-8'))
        data_type = "TEXT"
        if data_length > self.message_maximum_length * 1024:
            data_type = "ZIP"
            data = base64.b64encode(zlib.compress(data))

        message_attributes = {
            "Subject": {
                "DataType": "String",
                "StringValue": subject
            },
            "MessageType": {
                "DataType": "String",
                "StringValue": data_type
            },
            "FromStore": {
                "DataType": "String",
                "StringValue": self.store_code
            }
        }
        if self.context is not None:
            message_attributes["Context"] = {
                "DataType": "String",
                "StringValue": self.context
            }

        if self.server_queue_url is not None:
            response = self.sqs_client.send_message(
                QueueUrl=self.server_queue_url,
                MessageBody=data.decode("utf-8"),
                MessageAttributes=message_attributes,
            )
        else:
            response = self.sns_client.publish(
                TopicArn=self.topic_arn,
                Message=data,
                Subject=event.subject,
                MessageAttributes=message_attributes
            )

        if "MessageId" not in response or response["MessageId"] is None or response["MessageId"] == "":
            logger.info("MessageId not found on the response")

    def format_exception(self, message_bus, message, event, input_obj, exception):
        return

    def format_parse_exception(self, message_bus, message, event, data, exception):
        return



from production.command import ProductionCommand


class DevicesUpdatedCommand(ProductionCommand):
    def __init__(self):
        super(DevicesUpdatedCommand, self).__init__(None, None)

    def get_hash_extra_value(
        self,   # type: DevicesUpdatedCommand
    ):
        return "DevicesUpdatedCommand"




import json
import logging
import base64

from xml.etree import cElementTree as eTree
from bustoken import (
    TK_ELECTRONIC_PAYMENT_FINISH_PAYMENT,
    TK_ELECTRONIC_PAYMENT_NOT_ALL_PAYMENTS_REFUNDED,
    TK_ELECTRONIC_PAYMENT_REFUND_CANCELED
)
from msgbus import FM_PARAM, TK_SYS_ACK
from sysactions import (
    mbcontext,
    show_messagebox,
    StopAction
)

from actions.eft import (
    get_electronic_payment_component_name,
    get_electronic_tenders,
    has_electronic_payment,
    check_eft_is_valid,
)

logger = logging.getLogger("TableActions")


def finish_eft_transactions(
    pos_id,                 # type: str
    order,                  # type: eTree
    status,                 # type: str
    tender_id="",           # type: str
    is_table_service=False  # type: bool
):
    # type: (...) -> None
    from actions.config.is_tef_available import get_eft_processor_name

    totaled_states = order.findall("StateHistory/State[@state='TOTALED']")
    if len(totaled_states) == 0:
        return

    electronic_tenders = get_electronic_tenders(order)
    if len(electronic_tenders) == 0:
        return

    component_name = get_electronic_payment_component_name(pos_id=pos_id)
    if component_name is None:
        return

    eft_processor = get_eft_processor_name(pos_id=pos_id)
    if not check_eft_is_valid(eft_processor=eft_processor):
        return

    order_id = order.get("orderId")
    last_totaled = totaled_states[-1]
    last_totaled_date = last_totaled.get("timestamp").replace("-", "").replace(":", "")
    fiscal_date = last_totaled_date[:8]
    fiscal_time = last_totaled_date[9:15]
    status = status if has_electronic_payment(order) or is_table_service else "0"
    tender_details_str = base64.b64encode(json.dumps(electronic_tenders))
    data = ";".join([str(pos_id), order_id, fiscal_date, fiscal_time, status, tender_details_str, str(tender_id)])
    _finish_transaction(pos_id, order_id, status, data, component_name)


def _finish_transaction(
    pos_id,             # type: str
    order_id,           # type: str
    status,             # type: str
    data,               # type: str
    component_name,     # type: str
):
    # type: (...) -> None

    for try_num in range(1, 4):
        try:
            msg = "[finish_eft_transactions] Trying to send a finish message. Order #: {}; Try #: {}; STATUS #: {}"
            logger.info(msg.format(order_id, try_num, status))
            ten_minutes_timeout = 600000000
            ret = mbcontext.MB_EasySendMessage(
                dest_name=component_name,
                token=TK_ELECTRONIC_PAYMENT_FINISH_PAYMENT,
                format=FM_PARAM,
                data=data,
                timeout=ten_minutes_timeout
            )

            if ret.token == TK_SYS_ACK:
                logger.info("[finish_eft_transactions] Success finished eft transaction. Order #: {}".format(order_id))
                return
            elif ret.token == TK_ELECTRONIC_PAYMENT_NOT_ALL_PAYMENTS_REFUNDED:
                show_messagebox(pos_id, "$SOME_PAYMENTS_WERE_NOT_REFUNDED")
                continue
            elif ret.token == TK_ELECTRONIC_PAYMENT_REFUND_CANCELED:
                raise StopAction()

            message = "[finish_eft_transactions] ACK not received. Order #: {} - {} - {}"
            logger.info(message.format(order_id, ret.token, ret.data))
            _show_internal_error_message(pos_id, status)

        except Exception as ex:
            if isinstance(ex, StopAction):
                raise ex

            _show_internal_error_message(pos_id, status)

            msg = "[finish_eft_transactions] Error sending message. Order #: {} / Try #: {} / Status: {}"
            msg = msg.format(order_id, try_num, status)
            if try_num == 3:
                logger.exception(msg, exc_info=True)
            else:
                logger.info(msg)

    _show_manual_process_message(pos_id, status)


def _show_internal_error_message(
    pos_id,     # type: str
    status      # type: str
):
    # type: (...) -> None

    _display_message(
        pos_id,
        status,
        "$INTERNAL_ERROR_REFUNDING_ELECTRONIC_PAYMENTS",
        "$INTERNAL_ERROR_CONFIRMING_ELECTRONIC_PAYMENTS"
    )


def _show_manual_process_message(
    pos_id,     # type: str
    status      # type: str
):
    # type: (...) -> None

    _display_message(
        pos_id,
        status,
        "$ERROR_REFUNDING_ELECTRONIC_PAYMENTS_MANUAL_PROCESS_NEEDED",
        "$ERROR_CONFIRMING_ELECTRONIC_PAYMENTS_MANUAL_PROCESS_NEEDED"
    )


def _display_message(
    pos_id,             # type: str
    status,             # type: str
    error_refunding,    # type: str
    error_confirming    # type: str
):
    # type: (...) -> None

    message = error_refunding if status == "0" else error_confirming
    show_messagebox(pos_id, message)



"""
This module contains helper functions to ease the transition from ``pytz`` to
another :pep:`495`-compatible library.
"""
from . import _common, _compat
from ._impl import _PytzShimTimezone

_PYTZ_BASE_CLASSES = None


def is_pytz_zone(tz):
    """Check if a time zone is a ``pytz`` time zone.

    This will only import ``pytz`` if it has already been imported, and does
    not rely on the existence of the ``localize`` or ``normalize`` methods
    (since the shim classes also have these methods, but are not ``pytz``
    zones).
    """

    # If pytz is not in sys.modules, then we will assume the time zone is not a
    # pytz zone. It is possible that someone has manipulated sys.modules to
    # remove pytz, but that's the kind of thing that causes all kinds of other
    # problems anyway, so we'll call that an unsupported configuration.
    if not _common.pytz_imported():
        return False

    if _PYTZ_BASE_CLASSES is None:
        _populate_pytz_base_classes()

    return isinstance(tz, _PYTZ_BASE_CLASSES)


def upgrade_tzinfo(tz):
    """Convert a ``pytz`` or shim timezone into its modern equivalent.

    The shim classes are thin wrappers around :mod:`zoneinfo` or
    :mod:`dateutil.tz` implementations of the :class:`datetime.tzinfo` base
    class. This function removes the shim and returns the underlying "upgraded"
    time zone.

    When passed a ``pytz`` zone (not a shim), this returns the non-``pytz``
    equivalent. This may fail if ``pytz`` is using a data source incompatible
    with the upgraded provider's data source, or if the ``pytz`` zone was built
    from a file rather than an IANA key.

    When passed an object that is not a shim or a ``pytz`` zone, this returns
    the original object.

    :param tz:
        A :class:`datetime.tzinfo` object.

    :raises KeyError:
        If a ``pytz`` zone is passed to the function with no equivalent in the
        :pep:`495`-compatible library's version of the Olson database.

    :return:
        A :pep:`495`-compatible equivalent of any ``pytz`` or shim
        class, or the original object.
    """
    if isinstance(tz, _PytzShimTimezone):
        return tz._zone

    if is_pytz_zone(tz):
        if tz.zone is None:
            # This is a fixed offset zone
            offset = tz.utcoffset(None)
            offset_minutes = offset.total_seconds() / 60

            return _compat.get_fixed_offset_zone(offset_minutes)

        if tz.zone == "UTC":
            return _compat.UTC

        return _compat.get_timezone(tz.zone)

    return tz


def _populate_pytz_base_classes():
    import pytz
    from pytz.tzinfo import BaseTzInfo

    base_classes = (BaseTzInfo, pytz._FixedOffset)

    # In releases prior to 2018.4, pytz.UTC was not a subclass of BaseTzInfo
    if not isinstance(pytz.UTC, BaseTzInfo):  # pragma: nocover
        base_classes = base_classes + (type(pytz.UTC),)

    global _PYTZ_BASE_CLASSES
    _PYTZ_BASE_CLASSES = base_classes



from abc import (
    ABCMeta,
    abstractmethod,
)
from typing import Dict


class ProductRepository(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_order_items_names(
        self,
        order_items,    # type: str
    ):
        # type: (...) -> Dict

        raise NotImplementedError



import sys
import shutil


def shutil_rmtree(path, ignore_errors=False, onexc=None):
    if sys.version_info >= (3, 12):
        return shutil.rmtree(path, ignore_errors, onexc=onexc)

    def _handler(fn, path, excinfo):
        return onexc(fn, path, excinfo[1])

    return shutil.rmtree(path, ignore_errors, onerror=_handler)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from opentelemetry.context import create_key, get_value, set_value
from opentelemetry.context.context import Context
from opentelemetry.trace.span import INVALID_SPAN, Span

SPAN_KEY = "current-span"
_SPAN_KEY = create_key("current-span")


def set_span_in_context(
    span: Span, context: Optional[Context] = None
) -> Context:
    """Set the span in the given context.

    Args:
        span: The Span to set.
        context: a Context object. if one is not passed, the
            default current context is used instead.
    """
    ctx = set_value(_SPAN_KEY, span, context=context)
    return ctx


def get_current_span(context: Optional[Context] = None) -> Span:
    """Retrieve the current span.

    Args:
        context: A Context object. If one is not passed, the
            default current context is used instead.

    Returns:
        The Span set in the context if it exists. INVALID_SPAN otherwise.
    """
    span = get_value(_SPAN_KEY, context=context)
    if span is None or not isinstance(span, Span):
        return INVALID_SPAN
    return span



# -*- coding: utf-8 -*-

from json import JSONEncoder

from typing import Dict, Any

from ._Course import Course


class ProductionModelJsonEncoder(JSONEncoder):
    def default(self, o):
        instances_dictionary = {
            Course: self._get_course_json
        }

        if type(o) in instances_dictionary:
            return instances_dictionary[type(o)](o)

        return super(ProductionModelJsonEncoder, self).default(o)

    @staticmethod
    def _get_course_json(o):
        # type: (Course) -> Dict[str, Any]

        return {
            "id": o.course_id,
            "name": o.name,
            "description": o.description,
            "abbreviation": o.abbreviation,
            "delay": o.delay
        }



data = (
'ggyeols',    # 0x00
'ggyeolt',    # 0x01
'ggyeolp',    # 0x02
'ggyeolh',    # 0x03
'ggyeom',    # 0x04
'ggyeob',    # 0x05
'ggyeobs',    # 0x06
'ggyeos',    # 0x07
'ggyeoss',    # 0x08
'ggyeong',    # 0x09
'ggyeoj',    # 0x0a
'ggyeoc',    # 0x0b
'ggyeok',    # 0x0c
'ggyeot',    # 0x0d
'ggyeop',    # 0x0e
'ggyeoh',    # 0x0f
'ggye',    # 0x10
'ggyeg',    # 0x11
'ggyegg',    # 0x12
'ggyegs',    # 0x13
'ggyen',    # 0x14
'ggyenj',    # 0x15
'ggyenh',    # 0x16
'ggyed',    # 0x17
'ggyel',    # 0x18
'ggyelg',    # 0x19
'ggyelm',    # 0x1a
'ggyelb',    # 0x1b
'ggyels',    # 0x1c
'ggyelt',    # 0x1d
'ggyelp',    # 0x1e
'ggyelh',    # 0x1f
'ggyem',    # 0x20
'ggyeb',    # 0x21
'ggyebs',    # 0x22
'ggyes',    # 0x23
'ggyess',    # 0x24
'ggyeng',    # 0x25
'ggyej',    # 0x26
'ggyec',    # 0x27
'ggyek',    # 0x28
'ggyet',    # 0x29
'ggyep',    # 0x2a
'ggyeh',    # 0x2b
'ggo',    # 0x2c
'ggog',    # 0x2d
'ggogg',    # 0x2e
'ggogs',    # 0x2f
'ggon',    # 0x30
'ggonj',    # 0x31
'ggonh',    # 0x32
'ggod',    # 0x33
'ggol',    # 0x34
'ggolg',    # 0x35
'ggolm',    # 0x36
'ggolb',    # 0x37
'ggols',    # 0x38
'ggolt',    # 0x39
'ggolp',    # 0x3a
'ggolh',    # 0x3b
'ggom',    # 0x3c
'ggob',    # 0x3d
'ggobs',    # 0x3e
'ggos',    # 0x3f
'ggoss',    # 0x40
'ggong',    # 0x41
'ggoj',    # 0x42
'ggoc',    # 0x43
'ggok',    # 0x44
'ggot',    # 0x45
'ggop',    # 0x46
'ggoh',    # 0x47
'ggwa',    # 0x48
'ggwag',    # 0x49
'ggwagg',    # 0x4a
'ggwags',    # 0x4b
'ggwan',    # 0x4c
'ggwanj',    # 0x4d
'ggwanh',    # 0x4e
'ggwad',    # 0x4f
'ggwal',    # 0x50
'ggwalg',    # 0x51
'ggwalm',    # 0x52
'ggwalb',    # 0x53
'ggwals',    # 0x54
'ggwalt',    # 0x55
'ggwalp',    # 0x56
'ggwalh',    # 0x57
'ggwam',    # 0x58
'ggwab',    # 0x59
'ggwabs',    # 0x5a
'ggwas',    # 0x5b
'ggwass',    # 0x5c
'ggwang',    # 0x5d
'ggwaj',    # 0x5e
'ggwac',    # 0x5f
'ggwak',    # 0x60
'ggwat',    # 0x61
'ggwap',    # 0x62
'ggwah',    # 0x63
'ggwae',    # 0x64
'ggwaeg',    # 0x65
'ggwaegg',    # 0x66
'ggwaegs',    # 0x67
'ggwaen',    # 0x68
'ggwaenj',    # 0x69
'ggwaenh',    # 0x6a
'ggwaed',    # 0x6b
'ggwael',    # 0x6c
'ggwaelg',    # 0x6d
'ggwaelm',    # 0x6e
'ggwaelb',    # 0x6f
'ggwaels',    # 0x70
'ggwaelt',    # 0x71
'ggwaelp',    # 0x72
'ggwaelh',    # 0x73
'ggwaem',    # 0x74
'ggwaeb',    # 0x75
'ggwaebs',    # 0x76
'ggwaes',    # 0x77
'ggwaess',    # 0x78
'ggwaeng',    # 0x79
'ggwaej',    # 0x7a
'ggwaec',    # 0x7b
'ggwaek',    # 0x7c
'ggwaet',    # 0x7d
'ggwaep',    # 0x7e
'ggwaeh',    # 0x7f
'ggoe',    # 0x80
'ggoeg',    # 0x81
'ggoegg',    # 0x82
'ggoegs',    # 0x83
'ggoen',    # 0x84
'ggoenj',    # 0x85
'ggoenh',    # 0x86
'ggoed',    # 0x87
'ggoel',    # 0x88
'ggoelg',    # 0x89
'ggoelm',    # 0x8a
'ggoelb',    # 0x8b
'ggoels',    # 0x8c
'ggoelt',    # 0x8d
'ggoelp',    # 0x8e
'ggoelh',    # 0x8f
'ggoem',    # 0x90
'ggoeb',    # 0x91
'ggoebs',    # 0x92
'ggoes',    # 0x93
'ggoess',    # 0x94
'ggoeng',    # 0x95
'ggoej',    # 0x96
'ggoec',    # 0x97
'ggoek',    # 0x98
'ggoet',    # 0x99
'ggoep',    # 0x9a
'ggoeh',    # 0x9b
'ggyo',    # 0x9c
'ggyog',    # 0x9d
'ggyogg',    # 0x9e
'ggyogs',    # 0x9f
'ggyon',    # 0xa0
'ggyonj',    # 0xa1
'ggyonh',    # 0xa2
'ggyod',    # 0xa3
'ggyol',    # 0xa4
'ggyolg',    # 0xa5
'ggyolm',    # 0xa6
'ggyolb',    # 0xa7
'ggyols',    # 0xa8
'ggyolt',    # 0xa9
'ggyolp',    # 0xaa
'ggyolh',    # 0xab
'ggyom',    # 0xac
'ggyob',    # 0xad
'ggyobs',    # 0xae
'ggyos',    # 0xaf
'ggyoss',    # 0xb0
'ggyong',    # 0xb1
'ggyoj',    # 0xb2
'ggyoc',    # 0xb3
'ggyok',    # 0xb4
'ggyot',    # 0xb5
'ggyop',    # 0xb6
'ggyoh',    # 0xb7
'ggu',    # 0xb8
'ggug',    # 0xb9
'ggugg',    # 0xba
'ggugs',    # 0xbb
'ggun',    # 0xbc
'ggunj',    # 0xbd
'ggunh',    # 0xbe
'ggud',    # 0xbf
'ggul',    # 0xc0
'ggulg',    # 0xc1
'ggulm',    # 0xc2
'ggulb',    # 0xc3
'gguls',    # 0xc4
'ggult',    # 0xc5
'ggulp',    # 0xc6
'ggulh',    # 0xc7
'ggum',    # 0xc8
'ggub',    # 0xc9
'ggubs',    # 0xca
'ggus',    # 0xcb
'gguss',    # 0xcc
'ggung',    # 0xcd
'gguj',    # 0xce
'gguc',    # 0xcf
'gguk',    # 0xd0
'ggut',    # 0xd1
'ggup',    # 0xd2
'gguh',    # 0xd3
'ggweo',    # 0xd4
'ggweog',    # 0xd5
'ggweogg',    # 0xd6
'ggweogs',    # 0xd7
'ggweon',    # 0xd8
'ggweonj',    # 0xd9
'ggweonh',    # 0xda
'ggweod',    # 0xdb
'ggweol',    # 0xdc
'ggweolg',    # 0xdd
'ggweolm',    # 0xde
'ggweolb',    # 0xdf
'ggweols',    # 0xe0
'ggweolt',    # 0xe1
'ggweolp',    # 0xe2
'ggweolh',    # 0xe3
'ggweom',    # 0xe4
'ggweob',    # 0xe5
'ggweobs',    # 0xe6
'ggweos',    # 0xe7
'ggweoss',    # 0xe8
'ggweong',    # 0xe9
'ggweoj',    # 0xea
'ggweoc',    # 0xeb
'ggweok',    # 0xec
'ggweot',    # 0xed
'ggweop',    # 0xee
'ggweoh',    # 0xef
'ggwe',    # 0xf0
'ggweg',    # 0xf1
'ggwegg',    # 0xf2
'ggwegs',    # 0xf3
'ggwen',    # 0xf4
'ggwenj',    # 0xf5
'ggwenh',    # 0xf6
'ggwed',    # 0xf7
'ggwel',    # 0xf8
'ggwelg',    # 0xf9
'ggwelm',    # 0xfa
'ggwelb',    # 0xfb
'ggwels',    # 0xfc
'ggwelt',    # 0xfd
'ggwelp',    # 0xfe
'ggwelh',    # 0xff
)



# -*- coding: utf-8 -*-

import logging
from xml.etree.ElementTree import Element
from xml.etree import cElementTree as eTree

from msgbus import (
    TK_SLCTRL_OMGR_CLEARTENDERS,
    FM_PARAM,
    TK_SYS_ACK,
    MBMessage,
)

from actions import mb_context
from sysactions import (
    translate_message,
    get_model,
    send_message,
    get_posot,
    get_current_order,
)
from typing import (
    List,
    Optional,
    Tuple,
)

from application.repository import OrderRepository


class OrderInteractor(object):
    def __init__(self, pos_id):
        # type: (str) -> None

        self.logger = logging.getLogger("PosActions")
        self.order_repository = OrderRepository(mb_context)
        self.model = get_model(pos_id)
        self.pos_id = pos_id

    def get_orders_benefit(self, initial_date, end_date, operator_id):
        # type: (str, str, Optional[str]) -> List

        from msgbusboundary import MsgBusTableService as tableService
        try:
            all_coupons = translate_message(self.model, "ALL_COUPONS")
            pos_list = sorted(tableService().get_exclusive_pos_list())
            result = {'All': {'index': 0, 'id': None, 'name': all_coupons}}
            benefits = self.order_repository.get_orders_benefit(initial_date, end_date, operator_id, pos_list)
            index = 1
            for benefit in benefits:
                benefit_detail = benefit["benefit"]
                benefit_id = benefit_detail.get("id")
                if benefit_id in result.keys():
                    continue
                name = benefit_detail.get("userDescription").encode('utf-8')
                key = benefit_detail.get("name").encode('utf-8')
                data = {'index': index, 'id': benefit_id, 'name': name, 'key': key}
                result[benefit_id] = data
                index += 1

            return sorted(result.values(), key=lambda k: k['index'])
        except (Exception,):
            self.logger.exception("Error getting orders benefit")
            return []

    def get_fiscalized_orders(
            self,
            from_date,  # type: str
            pos_list,  # type: List[int]
    ):
        # type: (...) -> List

        return self.order_repository.get_fiscalized_orders(
            from_date=from_date,
            pos_list=pos_list,
        )

    def clear_order_tender(
        self,
        order_tender_id,            # type: str
        order_id,                   # type: str
        blk_notify="0",             # type: str
        tender_id="",               # type: str
        keep_tender_records="0",    # type: str
    ):
        # type: (...) -> None

        self.logger.info("Cleaning order tender. {}#{}. TenderId: {}".format(order_tender_id, order_id, tender_id))

        params = (
            blk_notify,
            self.pos_id,
            tender_id,
            order_tender_id,
            keep_tender_records,
            order_id
        )

        response = self.send_order_message(
            params=params,
            token=TK_SLCTRL_OMGR_CLEARTENDERS,
        )

        if response.token != TK_SYS_ACK:
            params = response.data.split('\0')
            self.logger.error("Error clear order tender: {} - {} - {}", params[0], params[0], params[1])

    def get_order_picture(
        self,
        pos_id=None,                # type: Optional[str]
        order_id="",                # type: Optional[str]
        use_order_id_model=True,    # type: Optional[bool]
    ):
        # type: (...) -> Optional[Element]

        pos_id = self.pos_id if pos_id is None else pos_id

        model = get_model(posid=pos_id)
        pos_ot = get_posot(model=model)

        if use_order_id_model is None:
            current_order = get_current_order(model=model)
            order_id = current_order.get("orderId")

        return eTree.XML(
            pos_ot.orderPicture(
                posid=pos_id,
                orderid=order_id,
            )
        )

    def get_order_from_order_picture(
        self,                           # type: OrderInteractor
        pos_id=None,                    # type: Optional[str]
        order_id="",                    # type: Optional[str]
        use_order_id_model=True,        # type: Optional[bool]
    ):
        # type: (...) -> Element

        order_picture = self.get_order_picture(
            pos_id=pos_id,
            order_id=order_id,
            use_order_id_model=use_order_id_model
        )
        return order_picture.find("Order") if order_picture.find("Order") else order_picture

    def send_order_message(
        self,
        params,     # type: Tuple
        token,      # type: int
    ):
        # type: (...) -> MBMessage

        params = "\0".join(str(p) for p in params)

        return send_message(
            dest_name="ORDERMGR%d" % int(self.pos_id),
            token=token,
            format=FM_PARAM,
            data=params,
            timeout=1000000 * 120
        )

    def get_pos_order_picture(
        self,  # type:  OrderInteractor
    ):
        order_interactor = OrderInteractor(
            pos_id=self.pos_id,
        )
        return order_interactor.get_order_from_order_picture(
            pos_id=self.pos_id
        )

    def is_order_in_progress(
        self,       # type:  OrderInteractor
    ):
        # type (...) -> bool

        current_order = self.get_pos_order_picture()
        return current_order.get("state") == "IN_PROGRESS"



from typing import Any

from ._CommandType import CommandType


class Command(object):
    def __init__(self, command_type, data):
        # type: (CommandType, Any) -> None
        self.type = command_type
        self.data = data
        self.order = None

    def add_order(self, order):
        self.order = order



from typing import (
    Dict,
    List,
)

from nreports_api.shared import JSON


class Line(Dict[str, JSON]):
    pass


class ReportDto(List[Line]):

    def __init__(
        self,   # type: ReportDto
        *args
    ):
        # type: (...) -> None
        super().__init__([Line(line) for line in args[0]])



from sysactions import show_messagebox, show_listbox, get_model, print_report, action, show_keyboard, StopAction

from application.repository import AccountRepository
from actions.util import (
    get_current_operator_session,
    is_current_operator_opened,
)
from actions.config import get_close_operator_conference_type
from actions.functions.dto import ConferenceTypes
from manager.reports import get_report_date, is_valid_date_range
from typing import Tuple

from .. import mb_context


@action
def show_operator_closing(pos_id, operator_id):
    from table_actions import considers_tef_on_bordereau

    end_date, operator_id, start_date = get_manual_inserted_data(operator_id, pos_id)

    model = get_model(pos_id)
    operator_current_session = ''
    if is_current_operator_opened(model):
        operator_current_session = get_current_operator_session(model=model)

    options, sessions = get_date_options_and_sessions(
        end_date,
        operator_id,
        pos_id,
        operator_current_session,
        start_date,
    )

    selected_date = show_listbox(pos_id, options, message="$SELECT_A_DATE")
    if selected_date is None:
        raise StopAction()

    session = sessions[selected_date]
    selected_date = session.split("period=")[1]

    report_name = "close_operator_report"
    bordereau_enabled = get_close_operator_conference_type() == ConferenceTypes.BORDEREAU.value
    print_report(
        pos_id,
        model,
        True,
        report_name,
        pos_id,
        selected_date,
        operator_id,
        session,
        "",
        bordereau_enabled,
        considers_tef_on_bordereau,
    )


def get_manual_inserted_data(
        operator_id,    # type: str
        pos_id          # type: str
):
    # type: (...) -> Tuple[str, str, str]

    start_date = get_report_date(pos_id, "$TYPE_INITIAL_DATE")
    if start_date is None:
        raise StopAction()

    end_date = get_report_date(pos_id, "$TYPE_FINAL_DATE")
    if end_date is None:
        raise StopAction()

    if is_valid_date_range(start_date, end_date, pos_id) is False:
        raise StopAction()

    if not operator_id:
        operator_id = show_keyboard(
            posid=pos_id,
            message="$TYPE_OPERATOR_ID_MESSAGE",
            mask="INTEGER",
            numpad=True,
            buttons='$OK|$PREVIOUS',
        )
        if operator_id is None:
            raise StopAction()

    return end_date, operator_id, start_date


def get_date_options_and_sessions(end_date, operator_id, pos_id, session, start_date):
    account_repository = AccountRepository(mb_context)
    sessions = account_repository.get_closing_operator_sessions(pos_id, operator_id, start_date, end_date)

    if session in sessions:
        sessions.remove(session)

    if not sessions:
        show_messagebox(pos_id, "$THERE_ARE_NO_DATA_TO_SHOW", "$INFORMATION")
        raise StopAction()

    options = []
    for current_session in sessions:
        session = current_session.split("period=")[1].split(",")[0]
        counter = current_session.split("count=")[1].split(",")[0]
        pos_type = 'QSR' if 'pos' in current_session else 'TS'
        formatted_option = "{}/{}/{} {} ({})".format(session[6:], session[4:6], session[:4], pos_type, counter)
        options.append(formatted_option)

    return options, sessions



# -*- coding: utf-8 -*-
try:
    from ._version import version as __version__
except ImportError:
    __version__ = 'unknown'

__all__ = ['easter', 'parser', 'relativedelta', 'rrule', 'tz',
           'utils', 'zoneinfo']



# tva.py - functions for handling Monacan TVA numbers
# coding: utf-8
#
# Copyright (C) 2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""n° TVA (taxe sur la valeur ajoutée, Monacan VAT number).

For VAT purposes Monaco is treated as territory of France. The number is
also validated the same as the French TVA, except that it is not based on
a French SIREN.

>>> compact('53 0000 04605')
'FR53000004605'
>>> validate('53 0000 04605')
'FR53000004605'
>>> validate('FR 61 954 506 077')  # French numbers are invalid
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""

from stdnum.exceptions import *
from stdnum.fr import tva


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return 'FR' + tva.compact(number)


def validate(number):
    """Check if the number is a valid VAT number. This checks the length,
    formatting and check digit."""
    number = tva.validate(number)
    if number[2:5] != '000':
        raise InvalidComponent()
    return 'FR' + number


def is_valid(number):
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



import json

from logging import Logger
from xml.etree import cElementTree as eTree
from util.xml2dict import convert_xml_to_dict
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Event
from sysactions import (
    get_model,
    get_posot,
)
from services.model import BOHOrderPictureRequest


class BOHOrderPictureInteractor(object):
    def __init__(
        self,
        message_bus,  # type: MbContextMessageBus
        logger,       # type: Logger
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.logger = logger
        self.event_action = "BohOrderPictureResponse"

    def process_request(
        self,
        request,  # type: BOHOrderPictureRequest
    ):
        # type: (...) -> None

        if request.pos_id is None or request.order_id is None:
            return

        pos_id = request.pos_id
        order_id = request.order_id
        model = get_model(posid=pos_id)
        pos_ot = get_posot(model=model)
        order_picture = eTree.XML(pos_ot.orderPicture(orderid=order_id)).find("Order")
        self.send_boh_order_picture_response_event(order_picture=order_picture)

    def send_boh_order_picture_response_event(
        self,
        order_picture,  # type: eTree.Element
    ):
        # type: (...) -> None

        try:
            event_data = {
                "orderPicture": convert_xml_to_dict(xml_object=order_picture),
            }
            event = Event(
                subject=self.event_action,
                event_type="",
                data=json.dumps(obj=event_data),
            )
            self.message_bus.publish_event(event=event)
            success_message = "BohOrderPictureResponse event sent successfully! order_id={0}"
            self.logger.info(msg=success_message.format(order_picture.get("orderId")))
        except Exception as _:  # noqa
            self.logger.exception(msg="Error sending BohOrderPictureResponse event")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: opentelemetry/proto/metrics/v1/metrics.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from opentelemetry.proto.common.v1 import common_pb2 as opentelemetry_dot_proto_dot_common_dot_v1_dot_common__pb2
from opentelemetry.proto.resource.v1 import resource_pb2 as opentelemetry_dot_proto_dot_resource_dot_v1_dot_resource__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,opentelemetry/proto/metrics/v1/metrics.proto\x12\x1eopentelemetry.proto.metrics.v1\x1a*opentelemetry/proto/common/v1/common.proto\x1a.opentelemetry/proto/resource/v1/resource.proto\"X\n\x0bMetricsData\x12I\n\x10resource_metrics\x18\x01 \x03(\x0b\x32/.opentelemetry.proto.metrics.v1.ResourceMetrics\"\xaf\x01\n\x0fResourceMetrics\x12;\n\x08resource\x18\x01 \x01(\x0b\x32).opentelemetry.proto.resource.v1.Resource\x12\x43\n\rscope_metrics\x18\x02 \x03(\x0b\x32,.opentelemetry.proto.metrics.v1.ScopeMetrics\x12\x12\n\nschema_url\x18\x03 \x01(\tJ\x06\x08\xe8\x07\x10\xe9\x07\"\x9f\x01\n\x0cScopeMetrics\x12\x42\n\x05scope\x18\x01 \x01(\x0b\x32\x33.opentelemetry.proto.common.v1.InstrumentationScope\x12\x37\n\x07metrics\x18\x02 \x03(\x0b\x32&.opentelemetry.proto.metrics.v1.Metric\x12\x12\n\nschema_url\x18\x03 \x01(\t\"\x92\x03\n\x06Metric\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0c\n\x04unit\x18\x03 \x01(\t\x12\x36\n\x05gauge\x18\x05 \x01(\x0b\x32%.opentelemetry.proto.metrics.v1.GaugeH\x00\x12\x32\n\x03sum\x18\x07 \x01(\x0b\x32#.opentelemetry.proto.metrics.v1.SumH\x00\x12>\n\thistogram\x18\t \x01(\x0b\x32).opentelemetry.proto.metrics.v1.HistogramH\x00\x12U\n\x15\x65xponential_histogram\x18\n \x01(\x0b\x32\x34.opentelemetry.proto.metrics.v1.ExponentialHistogramH\x00\x12:\n\x07summary\x18\x0b \x01(\x0b\x32\'.opentelemetry.proto.metrics.v1.SummaryH\x00\x42\x06\n\x04\x64\x61taJ\x04\x08\x04\x10\x05J\x04\x08\x06\x10\x07J\x04\x08\x08\x10\t\"M\n\x05Gauge\x12\x44\n\x0b\x64\x61ta_points\x18\x01 \x03(\x0b\x32/.opentelemetry.proto.metrics.v1.NumberDataPoint\"\xba\x01\n\x03Sum\x12\x44\n\x0b\x64\x61ta_points\x18\x01 \x03(\x0b\x32/.opentelemetry.proto.metrics.v1.NumberDataPoint\x12W\n\x17\x61ggregation_temporality\x18\x02 \x01(\x0e\x32\x36.opentelemetry.proto.metrics.v1.AggregationTemporality\x12\x14\n\x0cis_monotonic\x18\x03 \x01(\x08\"\xad\x01\n\tHistogram\x12G\n\x0b\x64\x61ta_points\x18\x01 \x03(\x0b\x32\x32.opentelemetry.proto.metrics.v1.HistogramDataPoint\x12W\n\x17\x61ggregation_temporality\x18\x02 \x01(\x0e\x32\x36.opentelemetry.proto.metrics.v1.AggregationTemporality\"\xc3\x01\n\x14\x45xponentialHistogram\x12R\n\x0b\x64\x61ta_points\x18\x01 \x03(\x0b\x32=.opentelemetry.proto.metrics.v1.ExponentialHistogramDataPoint\x12W\n\x17\x61ggregation_temporality\x18\x02 \x01(\x0e\x32\x36.opentelemetry.proto.metrics.v1.AggregationTemporality\"P\n\x07Summary\x12\x45\n\x0b\x64\x61ta_points\x18\x01 \x03(\x0b\x32\x30.opentelemetry.proto.metrics.v1.SummaryDataPoint\"\x86\x02\n\x0fNumberDataPoint\x12;\n\nattributes\x18\x07 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.KeyValue\x12\x1c\n\x14start_time_unix_nano\x18\x02 \x01(\x06\x12\x16\n\x0etime_unix_nano\x18\x03 \x01(\x06\x12\x13\n\tas_double\x18\x04 \x01(\x01H\x00\x12\x10\n\x06\x61s_int\x18\x06 \x01(\x10H\x00\x12;\n\texemplars\x18\x05 \x03(\x0b\x32(.opentelemetry.proto.metrics.v1.Exemplar\x12\r\n\x05\x66lags\x18\x08 \x01(\rB\x07\n\x05valueJ\x04\x08\x01\x10\x02\"\xe6\x02\n\x12HistogramDataPoint\x12;\n\nattributes\x18\t \x03(\x0b\x32\'.opentelemetry.proto.common.v1.KeyValue\x12\x1c\n\x14start_time_unix_nano\x18\x02 \x01(\x06\x12\x16\n\x0etime_unix_nano\x18\x03 \x01(\x06\x12\r\n\x05\x63ount\x18\x04 \x01(\x06\x12\x10\n\x03sum\x18\x05 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\rbucket_counts\x18\x06 \x03(\x06\x12\x17\n\x0f\x65xplicit_bounds\x18\x07 \x03(\x01\x12;\n\texemplars\x18\x08 \x03(\x0b\x32(.opentelemetry.proto.metrics.v1.Exemplar\x12\r\n\x05\x66lags\x18\n \x01(\r\x12\x10\n\x03min\x18\x0b \x01(\x01H\x01\x88\x01\x01\x12\x10\n\x03max\x18\x0c \x01(\x01H\x02\x88\x01\x01\x42\x06\n\x04_sumB\x06\n\x04_minB\x06\n\x04_maxJ\x04\x08\x01\x10\x02\"\xda\x04\n\x1d\x45xponentialHistogramDataPoint\x12;\n\nattributes\x18\x01 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.KeyValue\x12\x1c\n\x14start_time_unix_nano\x18\x02 \x01(\x06\x12\x16\n\x0etime_unix_nano\x18\x03 \x01(\x06\x12\r\n\x05\x63ount\x18\x04 \x01(\x06\x12\x10\n\x03sum\x18\x05 \x01(\x01H\x00\x88\x01\x01\x12\r\n\x05scale\x18\x06 \x01(\x11\x12\x12\n\nzero_count\x18\x07 \x01(\x06\x12W\n\x08positive\x18\x08 \x01(\x0b\x32\x45.opentelemetry.proto.metrics.v1.ExponentialHistogramDataPoint.Buckets\x12W\n\x08negative\x18\t \x01(\x0b\x32\x45.opentelemetry.proto.metrics.v1.ExponentialHistogramDataPoint.Buckets\x12\r\n\x05\x66lags\x18\n \x01(\r\x12;\n\texemplars\x18\x0b \x03(\x0b\x32(.opentelemetry.proto.metrics.v1.Exemplar\x12\x10\n\x03min\x18\x0c \x01(\x01H\x01\x88\x01\x01\x12\x10\n\x03max\x18\r \x01(\x01H\x02\x88\x01\x01\x12\x16\n\x0ezero_threshold\x18\x0e \x01(\x01\x1a\x30\n\x07\x42uckets\x12\x0e\n\x06offset\x18\x01 \x01(\x11\x12\x15\n\rbucket_counts\x18\x02 \x03(\x04\x42\x06\n\x04_sumB\x06\n\x04_minB\x06\n\x04_max\"\xc5\x02\n\x10SummaryDataPoint\x12;\n\nattributes\x18\x07 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.KeyValue\x12\x1c\n\x14start_time_unix_nano\x18\x02 \x01(\x06\x12\x16\n\x0etime_unix_nano\x18\x03 \x01(\x06\x12\r\n\x05\x63ount\x18\x04 \x01(\x06\x12\x0b\n\x03sum\x18\x05 \x01(\x01\x12Y\n\x0fquantile_values\x18\x06 \x03(\x0b\x32@.opentelemetry.proto.metrics.v1.SummaryDataPoint.ValueAtQuantile\x12\r\n\x05\x66lags\x18\x08 \x01(\r\x1a\x32\n\x0fValueAtQuantile\x12\x10\n\x08quantile\x18\x01 \x01(\x01\x12\r\n\x05value\x18\x02 \x01(\x01J\x04\x08\x01\x10\x02\"\xc1\x01\n\x08\x45xemplar\x12\x44\n\x13\x66iltered_attributes\x18\x07 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.KeyValue\x12\x16\n\x0etime_unix_nano\x18\x02 \x01(\x06\x12\x13\n\tas_double\x18\x03 \x01(\x01H\x00\x12\x10\n\x06\x61s_int\x18\x06 \x01(\x10H\x00\x12\x0f\n\x07span_id\x18\x04 \x01(\x0c\x12\x10\n\x08trace_id\x18\x05 \x01(\x0c\x42\x07\n\x05valueJ\x04\x08\x01\x10\x02*\x8c\x01\n\x16\x41ggregationTemporality\x12\'\n#AGGREGATION_TEMPORALITY_UNSPECIFIED\x10\x00\x12!\n\x1d\x41GGREGATION_TEMPORALITY_DELTA\x10\x01\x12&\n\"AGGREGATION_TEMPORALITY_CUMULATIVE\x10\x02*^\n\x0e\x44\x61taPointFlags\x12\x1f\n\x1b\x44\x41TA_POINT_FLAGS_DO_NOT_USE\x10\x00\x12+\n\'DATA_POINT_FLAGS_NO_RECORDED_VALUE_MASK\x10\x01\x42\x7f\n!io.opentelemetry.proto.metrics.v1B\x0cMetricsProtoP\x01Z)go.opentelemetry.io/proto/otlp/metrics/v1\xaa\x02\x1eOpenTelemetry.Proto.Metrics.V1b\x06proto3')

_AGGREGATIONTEMPORALITY = DESCRIPTOR.enum_types_by_name['AggregationTemporality']
AggregationTemporality = enum_type_wrapper.EnumTypeWrapper(_AGGREGATIONTEMPORALITY)
_DATAPOINTFLAGS = DESCRIPTOR.enum_types_by_name['DataPointFlags']
DataPointFlags = enum_type_wrapper.EnumTypeWrapper(_DATAPOINTFLAGS)
AGGREGATION_TEMPORALITY_UNSPECIFIED = 0
AGGREGATION_TEMPORALITY_DELTA = 1
AGGREGATION_TEMPORALITY_CUMULATIVE = 2
DATA_POINT_FLAGS_DO_NOT_USE = 0
DATA_POINT_FLAGS_NO_RECORDED_VALUE_MASK = 1


_METRICSDATA = DESCRIPTOR.message_types_by_name['MetricsData']
_RESOURCEMETRICS = DESCRIPTOR.message_types_by_name['ResourceMetrics']
_SCOPEMETRICS = DESCRIPTOR.message_types_by_name['ScopeMetrics']
_METRIC = DESCRIPTOR.message_types_by_name['Metric']
_GAUGE = DESCRIPTOR.message_types_by_name['Gauge']
_SUM = DESCRIPTOR.message_types_by_name['Sum']
_HISTOGRAM = DESCRIPTOR.message_types_by_name['Histogram']
_EXPONENTIALHISTOGRAM = DESCRIPTOR.message_types_by_name['ExponentialHistogram']
_SUMMARY = DESCRIPTOR.message_types_by_name['Summary']
_NUMBERDATAPOINT = DESCRIPTOR.message_types_by_name['NumberDataPoint']
_HISTOGRAMDATAPOINT = DESCRIPTOR.message_types_by_name['HistogramDataPoint']
_EXPONENTIALHISTOGRAMDATAPOINT = DESCRIPTOR.message_types_by_name['ExponentialHistogramDataPoint']
_EXPONENTIALHISTOGRAMDATAPOINT_BUCKETS = _EXPONENTIALHISTOGRAMDATAPOINT.nested_types_by_name['Buckets']
_SUMMARYDATAPOINT = DESCRIPTOR.message_types_by_name['SummaryDataPoint']
_SUMMARYDATAPOINT_VALUEATQUANTILE = _SUMMARYDATAPOINT.nested_types_by_name['ValueAtQuantile']
_EXEMPLAR = DESCRIPTOR.message_types_by_name['Exemplar']
MetricsData = _reflection.GeneratedProtocolMessageType('MetricsData', (_message.Message,), {
  'DESCRIPTOR' : _METRICSDATA,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.MetricsData)
  })
_sym_db.RegisterMessage(MetricsData)

ResourceMetrics = _reflection.GeneratedProtocolMessageType('ResourceMetrics', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEMETRICS,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.ResourceMetrics)
  })
_sym_db.RegisterMessage(ResourceMetrics)

ScopeMetrics = _reflection.GeneratedProtocolMessageType('ScopeMetrics', (_message.Message,), {
  'DESCRIPTOR' : _SCOPEMETRICS,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.ScopeMetrics)
  })
_sym_db.RegisterMessage(ScopeMetrics)

Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.Metric)
  })
_sym_db.RegisterMessage(Metric)

Gauge = _reflection.GeneratedProtocolMessageType('Gauge', (_message.Message,), {
  'DESCRIPTOR' : _GAUGE,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.Gauge)
  })
_sym_db.RegisterMessage(Gauge)

Sum = _reflection.GeneratedProtocolMessageType('Sum', (_message.Message,), {
  'DESCRIPTOR' : _SUM,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.Sum)
  })
_sym_db.RegisterMessage(Sum)

Histogram = _reflection.GeneratedProtocolMessageType('Histogram', (_message.Message,), {
  'DESCRIPTOR' : _HISTOGRAM,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.Histogram)
  })
_sym_db.RegisterMessage(Histogram)

ExponentialHistogram = _reflection.GeneratedProtocolMessageType('ExponentialHistogram', (_message.Message,), {
  'DESCRIPTOR' : _EXPONENTIALHISTOGRAM,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.ExponentialHistogram)
  })
_sym_db.RegisterMessage(ExponentialHistogram)

Summary = _reflection.GeneratedProtocolMessageType('Summary', (_message.Message,), {
  'DESCRIPTOR' : _SUMMARY,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.Summary)
  })
_sym_db.RegisterMessage(Summary)

NumberDataPoint = _reflection.GeneratedProtocolMessageType('NumberDataPoint', (_message.Message,), {
  'DESCRIPTOR' : _NUMBERDATAPOINT,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.NumberDataPoint)
  })
_sym_db.RegisterMessage(NumberDataPoint)

HistogramDataPoint = _reflection.GeneratedProtocolMessageType('HistogramDataPoint', (_message.Message,), {
  'DESCRIPTOR' : _HISTOGRAMDATAPOINT,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.HistogramDataPoint)
  })
_sym_db.RegisterMessage(HistogramDataPoint)

ExponentialHistogramDataPoint = _reflection.GeneratedProtocolMessageType('ExponentialHistogramDataPoint', (_message.Message,), {

  'Buckets' : _reflection.GeneratedProtocolMessageType('Buckets', (_message.Message,), {
    'DESCRIPTOR' : _EXPONENTIALHISTOGRAMDATAPOINT_BUCKETS,
    '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
    # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.ExponentialHistogramDataPoint.Buckets)
    })
  ,
  'DESCRIPTOR' : _EXPONENTIALHISTOGRAMDATAPOINT,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.ExponentialHistogramDataPoint)
  })
_sym_db.RegisterMessage(ExponentialHistogramDataPoint)
_sym_db.RegisterMessage(ExponentialHistogramDataPoint.Buckets)

SummaryDataPoint = _reflection.GeneratedProtocolMessageType('SummaryDataPoint', (_message.Message,), {

  'ValueAtQuantile' : _reflection.GeneratedProtocolMessageType('ValueAtQuantile', (_message.Message,), {
    'DESCRIPTOR' : _SUMMARYDATAPOINT_VALUEATQUANTILE,
    '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
    # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.SummaryDataPoint.ValueAtQuantile)
    })
  ,
  'DESCRIPTOR' : _SUMMARYDATAPOINT,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.SummaryDataPoint)
  })
_sym_db.RegisterMessage(SummaryDataPoint)
_sym_db.RegisterMessage(SummaryDataPoint.ValueAtQuantile)

Exemplar = _reflection.GeneratedProtocolMessageType('Exemplar', (_message.Message,), {
  'DESCRIPTOR' : _EXEMPLAR,
  '__module__' : 'opentelemetry.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opentelemetry.proto.metrics.v1.Exemplar)
  })
_sym_db.RegisterMessage(Exemplar)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n!io.opentelemetry.proto.metrics.v1B\014MetricsProtoP\001Z)go.opentelemetry.io/proto/otlp/metrics/v1\252\002\036OpenTelemetry.Proto.Metrics.V1'
  _AGGREGATIONTEMPORALITY._serialized_start=3487
  _AGGREGATIONTEMPORALITY._serialized_end=3627
  _DATAPOINTFLAGS._serialized_start=3629
  _DATAPOINTFLAGS._serialized_end=3723
  _METRICSDATA._serialized_start=172
  _METRICSDATA._serialized_end=260
  _RESOURCEMETRICS._serialized_start=263
  _RESOURCEMETRICS._serialized_end=438
  _SCOPEMETRICS._serialized_start=441
  _SCOPEMETRICS._serialized_end=600
  _METRIC._serialized_start=603
  _METRIC._serialized_end=1005
  _GAUGE._serialized_start=1007
  _GAUGE._serialized_end=1084
  _SUM._serialized_start=1087
  _SUM._serialized_end=1273
  _HISTOGRAM._serialized_start=1276
  _HISTOGRAM._serialized_end=1449
  _EXPONENTIALHISTOGRAM._serialized_start=1452
  _EXPONENTIALHISTOGRAM._serialized_end=1647
  _SUMMARY._serialized_start=1649
  _SUMMARY._serialized_end=1729
  _NUMBERDATAPOINT._serialized_start=1732
  _NUMBERDATAPOINT._serialized_end=1994
  _HISTOGRAMDATAPOINT._serialized_start=1997
  _HISTOGRAMDATAPOINT._serialized_end=2355
  _EXPONENTIALHISTOGRAMDATAPOINT._serialized_start=2358
  _EXPONENTIALHISTOGRAMDATAPOINT._serialized_end=2960
  _EXPONENTIALHISTOGRAMDATAPOINT_BUCKETS._serialized_start=2888
  _EXPONENTIALHISTOGRAMDATAPOINT_BUCKETS._serialized_end=2936
  _SUMMARYDATAPOINT._serialized_start=2963
  _SUMMARYDATAPOINT._serialized_end=3288
  _SUMMARYDATAPOINT_VALUEATQUANTILE._serialized_start=3232
  _SUMMARYDATAPOINT_VALUEATQUANTILE._serialized_end=3282
  _EXEMPLAR._serialized_start=3291
  _EXEMPLAR._serialized_end=3484
# @@protoc_insertion_point(module_scope)



from messagebus import AckMessage, NakMessage
from messageprocessorutil import UuidMessageProcessor
from production.manager import ProductionManager
from production.model.request import GetOrderXmlRequest


class GetOrderXmlProcessor(UuidMessageProcessor):
    def __init__(self, production_manager):
        # type: (ProductionManager) -> None

        super(GetOrderXmlProcessor, self).__init__()
        self.production_manager = production_manager

    def parse_data(self, data, event_type):
        return GetOrderXmlRequest(int(data))

    def call_business(self, request):
        # type: (GetOrderXmlRequest) -> str

        return self.production_manager.get_order_xml(request.order_id)

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_message(message, AckMessage(result.encode("utf-8")))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, NakMessage(str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, NakMessage(str(exception)))



# -*- coding: utf-8 -*-

import json
import time
from typing import (
    Dict,
    List,
    Optional,
)

from xml.etree import cElementTree as eTree
import xml.etree.ElementTree as ET

import persistence
import sysactions
from sysactions import (
    get_poslist,
    show_messagebox,
    format_amount,
    get_model,
)
from actions.models import GLAccount, TransferType, CloseOperator
from actions.util import (
    get_drawer_amount,
    list_ids_opened_users,
    get_all_tender_types,
)
from dto import ConferenceTypes
from actions.config import (
    get_close_operator_conference_type,
    considers_tef_on_bordereau,
)
from application.repository import AccountRepository
from msgbus import TK_ACCOUNT_TRANSFER, FM_PARAM, TK_SYS_ACK
from mw_helper import show_bordereau_dialog
from systools import sys_log_exception

from .. import mb_context
from .. import pos_config
from ..drawer.do_open_drawer import doOpenDrawer
from ..util import (
    get_drawer_initial_amount,
    get_operator_info,
    get_payments_amount_by_type,
)


skim_limit_min = 5
skim_limit_max = 30
sangria_levels = None


@sysactions.action
def do_transfer(
    pos_id,                     # type: str
    transfer_type,              # type: int
    operator_id,                # type: str
    authorizer=None,            # type: Optional[str]
    pos_to_close_user=None,     # type: Optional[str]
    transfer_values=None,       # type: Optional[Dict[str, float]]
):
    # type: (...) -> dict

    if pos_to_close_user is None:
        pos_to_close_user = pos_id

    global skim_limit_min, skim_limit_max, sangria_levels
    skim_limit_min = pos_config.skim_digit_limit['min']
    skim_limit_max = pos_config.skim_digit_limit['max']
    sangria_levels = pos_config.sangria_levels
    transfer_type = int(transfer_type)

    model = sysactions.get_model(
        posid=pos_to_close_user,
    )
    _check_if_can_do_transfer(
        pos_id=pos_id,
        model=model,
        operator_id=operator_id,
        transfer_values=transfer_values,
    )

    _, session = get_operator_info(model, operator_id)
    period = sysactions.get_business_period(model)
    manager_id = sysactions.get_custom(model, 'Last Manager ID') or ""
    if not manager_id and authorizer is not None:
        manager_id = authorizer

    if transfer_type == TransferType.TRANSFER_CASH_IN or transfer_type == TransferType.TRANSFER_CASH_OUT:
        drawer_amount = get_drawer_amount(
            pos_id=pos_to_close_user,
            period=period,
            session=session,
        )
        return _process_cash_in_and_out(
            pos_id=pos_id,
            model=model,
            session=session,
            operator_id=operator_id,
            period=period,
            manager_id=manager_id,
            drawer_amount=drawer_amount,
            transfer_type=transfer_type,
            pos_to_close_user=pos_to_close_user,
        )
    else:
        sold_amount_by_tender_type = get_sold_amount_by_tender_type(session)
        return _process_close_operator(
            pos_id=pos_id,
            session=session,
            period=period,
            manager_id=manager_id,
            amount_by_tender_type=sold_amount_by_tender_type,
            pos_to_close_user=pos_to_close_user,
            transfer_values=transfer_values,
        )


def get_transfer_values(
    pos_id,     # type: str
    period,     # type: str
    session     # type: str
):
    # type: (...) -> List[Dict[str, float]]

    conference_type = get_close_operator_conference_type()
    has_valid_tender = _has_valid_bordereau_tender()
    if conference_type == ConferenceTypes.BORDEREAU.value and has_valid_tender:
        return _get_tender_types_inserted_values(pos_id=pos_id)

    if conference_type == ConferenceTypes.CASH_DRAWER.value and has_valid_tender:
        message = "$TYPE_THE_DRAWER_VALUE"
        amount = sysactions.show_keyboard(
            posid=pos_id,
            message=message,
            title='',
            mask="CURRENCY",
            numpad=True,
            timeout=720000,
        )
        if amount is None:
            raise sysactions.StopAction()

        return [{"tenderId": 0, "amount": amount}]

    drawer_amount = get_drawer_amount(
        pos_id=pos_id,
        period=period,
        session=session,
    )
    return [{"tenderId": 0, "amount": drawer_amount}]


def _show_drawer_funds_message(
    pos_id,                 # type: str
    initial_value_drawer,   # type: float
    envelope_amount,        # type: float
    transfer_amount,        # type: float
):
    # type: (...) -> None

    model = get_model(posid=pos_id)
    message = "$DRAWER_FUNDS_MESSAGE|{0}|{1}".format(
        format_amount(
            model=model,
            amount=initial_value_drawer,
        ),
        format_amount(
            model=model,
            amount=envelope_amount,
        )
    )

    if transfer_amount <= initial_value_drawer:
        message = "$DRAWER_RESERVE_FUNDS_MESSAGE|{0}".format(
            format_amount(
                model=model,
                amount=initial_value_drawer,
            )
        )

    sysactions.show_messagebox(
        posid=pos_id,
        message=message,
        timeout=720000,
        buttons="$OK",
    )


def _has_valid_bordereau_tender():
    # type: (...) -> bool

    for tender in get_all_tender_types():
        has_child = tender.get("hasChild")
        if has_child:
            continue

        is_electronic_payment = tender.get("electronicTypeId") != 0
        if is_electronic_payment and not considers_tef_on_bordereau():
            continue

        custom_attr = tender.get("customAttr")
        if not custom_attr:
            return True

        custom_attr = json.loads(custom_attr)
        show_in_bordereau = custom_attr.get("showInBordereau", True)
        if show_in_bordereau:
            return True

    return False


def _process_close_operator(
    pos_id,                     # type: str
    session,                    # type: str
    period,                     # type: str
    manager_id,                 # type: str
    amount_by_tender_type,      # type: Dict[int, List[Dict]]
    pos_to_close_user=None,     # type: str
    transfer_values=None,       # type: Optional[List[Dict[str, float]]]
):
    # type: (...) -> dict

    if transfer_values is None:
        transfer_values = get_transfer_values(
            pos_id=pos_id,
            period=period,
            session=session,
        )

    initial_value_drawer = _get_initial_drawer_amount(
        pos_id=pos_to_close_user,
        session=session,
    )
    transfer_amount = _get_cash_amount_in_transfer_values(transfer_values=transfer_values)

    envelope_amount = transfer_amount - initial_value_drawer
    envelope = _get_envelope(
        pos_id=pos_id,
        envelope_amount=envelope_amount,
    )

    _fill_tender_attributes(inserted_values=transfer_values)

    if pos_to_close_user is None:
        pos_to_close_user = pos_id

    try:
        operator = CloseOperator(
            pos_id=pos_to_close_user,
            values=transfer_values,
            envelope=envelope,
            session=session,
            payments=amount_by_tender_type,
            manager_id=manager_id,
            period=period,
        )
        has_some_difference = save_inserted_values(operator)
        gl_account = GLAccount(manager=manager_id)
        extra_info = json.dumps(gl_account, default=lambda o: o.__dict__, sort_keys=True)
        status = TransferType.FINAL_AMOUNT
        if initial_value_drawer:
            _show_drawer_funds_message(
                pos_id=pos_id,
                initial_value_drawer=initial_value_drawer,
                envelope_amount=envelope_amount,
                transfer_amount=transfer_amount,
            )

        data = "{}\0{}\0{}\0{}\0{}\0{}".format(
            session,
            status,
            "FINAL_AMOUNT",
            initial_value_drawer,
            0,
            extra_info
        )
        sysactions.send_message(
            dest_name="account%s" % pos_id,
            token=TK_ACCOUNT_TRANSFER,
            format=FM_PARAM,
            data=data,
        )

        if transfer_amount > 0:
            _send_audit_cash_event(
                xml_name="DeclaredAmount",
                event_name="DECLARED_AMOUNT",
                store_id=pos_config.store_id,
                period=period,
                pos_id=pos_id,
                session=session,
                transfer_amount=str(transfer_amount),
                timestamp=str(time.time()),
                extra_info=extra_info,
            )

        if has_some_difference:
            return {'need_ask_justification': True}

    except (Exception,):
        sysactions.show_messagebox(pos_id, message="$ERROR_SAVING_CLOSE_OPERATOR_REPORT")
        raise sysactions.StopAction()

    return {'need_ask_justification': False}


def _send_audit_cash_event(
    xml_name,           # type: str
    event_name,         # type: str
    store_id,           # type: str
    period,             # type: str
    pos_id,             # type: str
    session,            # type: str
    transfer_amount,    # type: str
    timestamp,          # type: str
    extra_info,         # type: str
):
    # type: (...) -> None

    declared_amount_xml = ET.Element(
        xml_name,
        storeId=store_id,
        period=period,
        posId=pos_id,
        sessionId=session,
        amount=transfer_amount,
        tenderId="0",
        timestamp=timestamp,
        extraInfo=extra_info,
    )

    mb_context.MB_EasyEvtSend(
        subject=event_name,
        type=event_name,
        xml=eTree.tostring(declared_amount_xml),
    )


def _get_cash_amount_in_transfer_values(
    transfer_values,     # type: Optional[List[Dict[str, float]]]
):
    # type: (...) -> float

    cash_transfer_amount = 0.0
    for value in transfer_values:
        if value.get("tenderId") != 0:
            continue
        cash_transfer_amount += float(value.get("amount"))

    return cash_transfer_amount


def _get_envelope(
    pos_id,             # type: str
    envelope_amount,    # type: float
):
    # type: (...) -> Optional[str]

    envelope = ""
    if envelope_amount > 0:
        envelope = _get_skim_number(pos_id)

    if envelope is None:
        raise sysactions.StopAction()

    return envelope


def save_inserted_values(
    operator,  # type: CloseOperator
):
    # type: (...) -> bool

    account_repository = AccountRepository(mb_context)
    total_bordereau_difference = account_repository.insert_close_operator_info(operator)
    total_bordereau_difference = abs(total_bordereau_difference)
    if total_bordereau_difference == 0:
        return False

    return total_bordereau_difference >= pos_config.min_value_to_ask_bordereau_justify_config


def get_justification(
    account_repository,    # type: AccountRepository
    pos_id,                # type: str
    session,               # type: str
):
    # type: (...) -> None

    justification = _get_justification(pos_id=pos_id)
    account_repository.update_justification(
        pos_id=pos_id,
        session=session,
        justification=justification,
    )


def _get_justification(
    pos_id,                                  # type: str
    dialog_message="$CASH_BREAK_MESSAGE",    # type: Optional[str]
    buttons="$OK",                           # type: Optional[str]
):
    # type: (...) -> str

    justification = None
    while justification is None or len(justification) < 5:
        justification = sysactions.show_keyboard(
            posid=pos_id,
            message=dialog_message,
            title="",
            timeout=1800000,
            buttons=buttons,
        )

        if justification is not None:
            justification = justification.strip()
            if len(justification) < 5:
                show_messagebox(
                    posid=pos_id,
                    message="$MINIMUM_5_CHARACTERS_REQUIRED",
                )
            else:
                break
        else:
            break
    return justification


def _process_cash_in_and_out(
        pos_id,                     # type: str
        model,                      # type: eTree.XML
        session,                    # type: str
        operator_id,                # type: str
        period,                     # type: str
        manager_id,                 # type: str
        drawer_amount,              # type: float
        transfer_type,              # type: int
        pos_to_close_user=None,     # type: str
):
    # type: (...) -> dict

    if pos_to_close_user is None:
        pos_to_close_user = pos_id

    tile = "$CASH_IN" if transfer_type == TransferType.TRANSFER_CASH_IN else "$CASH_OUT"
    transfer_text = "TRANSFER_CASH_IN" if transfer_type == TransferType.TRANSFER_CASH_IN else "TRANSFER_CASH_OUT"
    translated_transfer_text = sysactions.translate_message(model, transfer_text)
    message = "$ENTER_AMOUNT_TO_TRANSFER_TYPE|%s" % translated_transfer_text

    amount = _get_inserted_amount(
        pos_id=pos_to_close_user,
        message=message,
        title=tile,
    )
    if amount is None or amount == 0:
        raise sysactions.StopAction()

    if not is_valid_amount_by_transfer_type(
        pos_id=pos_to_close_user,
        transfer_type=transfer_type,
        amount=amount,
        drawer_amount=drawer_amount,
    ):
        raise sysactions.StopAction()

    envelope_number = ""
    justification = ""
    if transfer_type == TransferType.TRANSFER_CASH_OUT:
        if _get_ask_cash_out_justification():
            justification = _get_justification(
                pos_id=pos_id,
                dialog_message='$CASH_OUT_JUSTIFICATION',
                buttons="$OK|$CANCEL",
            )
        envelope_number = _get_skim_number(
            pos_id=pos_to_close_user,
        )
        if envelope_number is None or justification is None:
            raise sysactions.StopAction()

    return _open_drawer(
        pos_id=pos_id,
        model=model,
        operator_id=operator_id,
        manager_id=manager_id,
        period=period,
        envelope_number=envelope_number,
        justification=justification,
        session=session,
        transfer_type=transfer_type,
        transfer_text=transfer_text,
        amount=amount,
        pos_to_close_user=pos_to_close_user,
    )


def _open_drawer(
    pos_id,                 # type: str
    model,                  # type: eTree.XML
    operator_id,            # type: str
    manager_id,             # type: str
    period,                 # type: str
    envelope_number,        # type: str
    justification,          # type: str
    session,                # type: str
    transfer_type,          # type: int
    transfer_text,          # type: str
    amount,                 # type: float
    pos_to_close_user,      # type: str
):
    # type: (...) -> Optional[Dict[str, bool]]

    from table_actions import print_operator_copy

    gl_account = GLAccount(
        manager=manager_id,
        envelope=envelope_number,
        justification=justification,
    )
    extra_info = json.dumps(gl_account, default=lambda o: o.__dict__, sort_keys=True)
    data = "%s\0%s\0%s\0%s\0%s\0%s" % (session, transfer_type, transfer_text, amount, 0, extra_info)
    msg = sysactions.send_message(
        dest_name="account%s" % pos_to_close_user,
        token=TK_ACCOUNT_TRANSFER,
        format=FM_PARAM,
        data=data,
    )
    if msg.token == TK_SYS_ACK:
        doOpenDrawer(
            pos_id=pos_to_close_user,
            need_logged=operator_id != manager_id,
        )
        if sysactions.print_report(pos_id, model, True, "transfer_report", pos_id, operator_id, transfer_type, amount,
                                   period, envelope_number, False, justification):

            if print_operator_copy:
                sysactions.print_report(pos_id, model, False, "transfer_report", pos_id, operator_id, transfer_type,
                                        amount, period, envelope_number, True, justification)
            sysactions.show_messagebox(pos_id, "$OPERATION_SUCCEEDED")

        _send_audit_cash_event(
            xml_name="TransferInfo",
            event_name="TRANSFER_CASH_OUT",
            store_id=pos_config.store_id,
            period=period,
            pos_id=pos_id,
            session=session,
            transfer_amount=str(amount),
            timestamp=str(time.time()),
            extra_info=extra_info,
        )

        return {'need_ask_justification': False}

    sysactions.show_messagebox(pos_id, "$OPERATION_FAILED", "$ERROR")
    raise sysactions.StopAction()


def get_sold_amount_by_tender_type(
    session_id,     # type: str
):
    # type: (...) -> Dict[int, List[Dict]]

    pos_list = get_poslist()
    payments_amount = {}
    for pos_id in pos_list:
        get_payments_amount_by_type(
            pos_id,
            session_id,
            payments_amount,
        )

    return payments_amount


def valid_amount_by_skim_limit(amount):
    if pos_config.max_transfer_value == 0:
        return True
    if round(float(amount), 2) > round(pos_config.max_transfer_value, 2):
        return False
    return True


def _check_if_can_do_transfer(
    pos_id,             # type: str
    model,              # type: eTree.XML
    operator_id,        # type: str
    transfer_values,    # type: Optional[Dict[int, List[Dict]]]
):
    # type: (...) -> None

    if not sysactions.is_day_opened(model):
        sysactions.show_messagebox(pos_id, message="$DAY_IS_CLOSED")
        raise sysactions.StopAction()

    if operator_id == '':
        sysactions.show_messagebox(pos_id, message="$OPERATOR_IS_NOT_OPENED")
        raise sysactions.StopAction()

    if transfer_values is not None:
        return

    has_open_operator = _has_operator_opened(
        model=model,
        operator_id=operator_id,
    )
    if has_open_operator:
        return

    sysactions.show_messagebox(pos_id, message="$OPERATOR_IS_NOT_OPENED")
    raise sysactions.StopAction()


def _has_operator_opened(
    model,          # type: eTree.XML
    operator_id,    # type: str
):
    # type: (...) -> bool

    if int(operator_id) in list_ids_opened_users():
        return True

    login_type = model.find("WorkingMode").get("usrCtrlType")
    operators = model.findall(".//Operators/Operator")
    is_any_available = any([op.get("state") == "LOGGEDIN" for op in operators])
    if is_any_available and login_type == "TS":
        return True

    is_any_paused = any([op.get("state") == "PAUSED" for op in operators])
    if login_type == "QSR" and is_any_paused:
        return True

    return False


def _get_operator(model):
    return model.findall('.//Operator')


def is_valid_amount_by_transfer_type(pos_id, transfer_type, amount, drawer_amount):
    if not _inserted_amount_is_valid(amount):
        message = "$CASH_OUT_MAX_VALUE|%s" % "{0:.2f}".format(pos_config.max_transfer_value)
        sysactions.show_messagebox(pos_id, message, "$INFORMATION")
        return False

    is_cash_out = transfer_type == TransferType.TRANSFER_CASH_OUT
    if is_cash_out and drawer_amount < amount and pos_config.max_transfer_value != 0:
        sysactions.show_messagebox(pos_id, "$SKIM_VALUE_MUCH_BIGGER", "$INFORMATION")
        return False

    return True


def _inserted_amount_is_valid(amount):
    invalid_amount = amount in ("", ".") or round(float(amount), 2) < 0
    return not invalid_amount and valid_amount_by_skim_limit(amount)


def _get_skim_number(pos_id):
    if pos_config.auto_generate_skim_number:
        return _auto_generate_skim_number(pos_id)

    return _request_skim_number(pos_id)


def _get_ask_cash_out_justification():
    # type: (...) -> bool

    return pos_config.ask_cash_out_justification


def _request_skim_number(pos_id):
    valid_envelope_number = False
    skim_number = None
    while not valid_envelope_number:
        skim_number = sysactions.show_keyboard(pos_id, "$TYPE_SKIM_NUMBER", numpad=pos_config.keyboard_skim_numeric)
        if skim_number is not None and (len(skim_number) < skim_limit_min or len(skim_number) > skim_limit_max):
            msg = "$INVALID_SKIM_NUMBER|{0}|{1}".format(skim_limit_min, skim_limit_max)
            sysactions.show_messagebox(pos_id, msg)
            continue
        elif skim_number is None:
            return None
        valid_envelope_number = True
    return skim_number


def _auto_generate_skim_number(pos_id):
    model = sysactions.get_model(pos_id)
    period = sysactions.get_business_period(model)
    day = int(period[6:8] + period[4:6] + period[2:4])
    count_cash_out = _get_number_of_cash_out(pos_id, period)
    skim_number = "{:05d}{:06d}{}".format(abs(hash(pos_config.store_id)), day, count_cash_out)

    return skim_number


def _get_initial_drawer_amount(pos_id, session):
    return get_drawer_initial_amount(pos_id, session)


def _get_inserted_amount(
    pos_id,
    message,
    title="",
):
    amount = sysactions.show_keyboard(
        posid=pos_id,
        message=message,
        title=title,
        mask="CURRENCY",
        numpad=True,
        timeout=720000,
        buttons='$OK|$PREVIOUS',
    )
    return None if amount is None else round(float(amount), 2)


def _get_tender_types_inserted_values(pos_id):
    values = show_bordereau_dialog(pos_id)
    if values is None:
        raise sysactions.StopAction()

    return json.loads(values)


def _get_number_of_cash_out(pos_id, period):
    conn = None
    skim_count = 0
    try:
        conn = persistence.Driver().open(mb_context, pos_id)
        cursor = conn.select(
            """
                SELECT
                    COUNT(*)
                FROM account.Transfer
                WHERE
                    Description in ('FINAL_AMOUNT', 'TRANSFER_CASH_OUT')
                    AND Period = '{}'
                GROUP BY Period
            """.format(period)
        )
        for row in cursor:
            skim_count = int(row.get_entry(0) or 0)
        return skim_count

    except (Exception,):
        sys_log_exception("get amount of cash out Error")
    finally:
        if conn:
            conn.close()


def _fill_tender_attributes(
    inserted_values,  # type: List[Dict]
):
    # type: (...) -> None

    conn = None

    _normalize_amount_type(inserted_values)
    tender_ids = [str(value["tenderId"]) for value in inserted_values]

    try:
        conn = persistence.Driver().open(mb_context)
        cursor = conn.select(
            """
            SELECT TenderId, CustomAttr
            FROM TenderType
            WHERE TenderId IN ({})
            """.format(",".join(tender_ids))
        )
        for row in cursor:
            tender_id = int(row.get_entry("TenderId"))
            for value in inserted_values:
                if value["tenderId"] == tender_id:
                    custom_attr = json.loads(row.get_entry("CustomAttr") or "{}")
                    value["linkWithTef"] = str(custom_attr.get("linkWithTef", "false")).lower() == "true"
                    value["customAttr"] = custom_attr
                    break

    except:  # noqa
        sys_log_exception("Error filling tender attributes")
    finally:
        if conn:
            conn.close()


def _normalize_amount_type(
    inserted_values,  # type: List[Dict]
):
    # type: (...) -> None

    for value in inserted_values:
        if not isinstance(value["amount"], float):
            value["amount"] = float(value["amount"])



from messagebus import (
    Message,
    DefaultToken,
    DataType,
    Event,
    MessageBus,
)
from messageprocessorutil import UuidMessageProcessor
from application.interactor import HasPendingLoaderUpdateInteractor
from typing import (
    Any,
    Optional,
)


class HasPendingLoaderUpdateProcessor(UuidMessageProcessor):

    def __init__(
        self,           # type: HasPendingLoaderUpdateProcessor
        interactor,     # type: HasPendingLoaderUpdateInteractor
    ):
        # type: (...) -> None
        super(HasPendingLoaderUpdateProcessor, self).__init__()

        self.interactor = interactor

    def parse_data(
        self,           # type: HasPendingLoaderUpdateProcessor
        _,              # type: str
        event_type,     # type: str
    ):
        # type: (...) -> None

        return

    def call_business(
        self,   # type: HasPendingLoaderUpdateProcessor
        _,      # type: str
    ):
        # type: (...) -> bool

        return self.interactor.get_loader_pending_update()

    def format_response(
        self,           # type: HasPendingLoaderUpdateProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Event]
        request,        # type: Any
        result,         # type: Any
    ):
        # type: (...) -> None

        reply_data = {"loader_has_pending_update": result}

        reply_message = Message(
            token=DefaultToken.TK_SYS_ACK,
            data_type=DataType.param,
            data=reply_data,
        )
        message_bus.reply_message(
            message=message,
            reply_message=reply_message,
        )

    def format_exception(
        self,           # type: HasPendingLoaderUpdateProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Event]
        input_obj,      # type: None
        exception,      # type: Exception
    ):
        # type: (...) -> None

        exception_name = str(type(exception).__name__)
        message_bus.reply_message(
            message=message,
            reply_message=Message(
                token=DefaultToken.TK_SYS_NAK,
                data_type=DataType.param,
                data=exception_name,
            )
        )

    def format_parse_exception(
        self,           # type: HasPendingLoaderUpdateProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        data,           # type: str
        exception,      # type: Exception
    ):
        # type: (...) -> None

        return



class ToggleTagLineRequest(object):
    def __init__(self, order_id, view, lines, tag_name):
        self.order_id = order_id
        self.view = view
        self.lines = lines
        self.tag_name = tag_name



"""distutils.command.check

Implements the Distutils 'check' command.
"""

import contextlib

from ..core import Command
from ..errors import DistutilsSetupError

with contextlib.suppress(ImportError):
    import docutils.frontend
    import docutils.nodes
    import docutils.parsers.rst
    import docutils.utils

    class SilentReporter(docutils.utils.Reporter):
        def __init__(
            self,
            source,
            report_level,
            halt_level,
            stream=None,
            debug=0,
            encoding='ascii',
            error_handler='replace',
        ):
            self.messages = []
            super().__init__(
                source, report_level, halt_level, stream, debug, encoding, error_handler
            )

        def system_message(self, level, message, *children, **kwargs):
            self.messages.append((level, message, children, kwargs))
            return docutils.nodes.system_message(
                message, *children, level=level, type=self.levels[level], **kwargs
            )


class check(Command):
    """This command checks the meta-data of the package."""

    description = "perform some checks on the package"
    user_options = [
        ('metadata', 'm', 'Verify meta-data'),
        (
            'restructuredtext',
            'r',
            (
                'Checks if long string meta-data syntax '
                'are reStructuredText-compliant'
            ),
        ),
        ('strict', 's', 'Will exit with an error if a check fails'),
    ]

    boolean_options = ['metadata', 'restructuredtext', 'strict']

    def initialize_options(self):
        """Sets default values for options."""
        self.restructuredtext = 0
        self.metadata = 1
        self.strict = 0
        self._warnings = 0

    def finalize_options(self):
        pass

    def warn(self, msg):
        """Counts the number of warnings that occurs."""
        self._warnings += 1
        return Command.warn(self, msg)

    def run(self):
        """Runs the command."""
        # perform the various tests
        if self.metadata:
            self.check_metadata()
        if self.restructuredtext:
            if 'docutils' in globals():
                try:
                    self.check_restructuredtext()
                except TypeError as exc:
                    raise DistutilsSetupError(str(exc))
            elif self.strict:
                raise DistutilsSetupError('The docutils package is needed.')

        # let's raise an error in strict mode, if we have at least
        # one warning
        if self.strict and self._warnings > 0:
            raise DistutilsSetupError('Please correct your package.')

    def check_metadata(self):
        """Ensures that all required elements of meta-data are supplied.

        Required fields:
            name, version

        Warns if any are missing.
        """
        metadata = self.distribution.metadata

        missing = []
        for attr in 'name', 'version':
            if not getattr(metadata, attr, None):
                missing.append(attr)

        if missing:
            self.warn("missing required meta-data: %s" % ', '.join(missing))

    def check_restructuredtext(self):
        """Checks if the long string fields are reST-compliant."""
        data = self.distribution.get_long_description()
        for warning in self._check_rst_data(data):
            line = warning[-1].get('line')
            if line is None:
                warning = warning[1]
            else:
                warning = f'{warning[1]} (line {line})'
            self.warn(warning)

    def _check_rst_data(self, data):
        """Returns warnings when the provided data doesn't compile."""
        # the include and csv_table directives need this to be a path
        source_path = self.distribution.script_name or 'setup.py'
        parser = docutils.parsers.rst.Parser()
        settings = docutils.frontend.OptionParser(
            components=(docutils.parsers.rst.Parser,)
        ).get_default_values()
        settings.tab_width = 4
        settings.pep_references = None
        settings.rfc_references = None
        reporter = SilentReporter(
            source_path,
            settings.report_level,
            settings.halt_level,
            stream=settings.warning_stream,
            debug=settings.debug,
            encoding=settings.error_encoding,
            error_handler=settings.error_encoding_error_handler,
        )

        document = docutils.nodes.document(settings, reporter, source=source_path)
        document.note_source(source_path, -1)
        try:
            parser.parse(data, document)
        except AttributeError as e:
            reporter.messages.append((
                -1,
                'Could not finish the parsing: %s.' % e,
                '',
                {},
            ))

        return reporter.messages



import json
from typing import Optional

from messagebus import (
    DataType,
    DefaultToken,
    Message,
    MessageBus,
)
from .shared import error
from .shared.exception import NReportsException
from .shared.reports import Reports
from .shared.token import TK_NREPORTS_BUILD_REPORT


class NReports(object):

    def __init__(
        self,           # type: NReports
        message_bus,    # type: MessageBus
    ):
        # type: (...) -> None

        self.message_bus = message_bus

    def request_report(
        self,           # type: NReports
        pos_id,         # type: int
        report_name,    # type: str
        **kwargs
    ):
        # type: (...) -> str

        request = self._build_request_data(
            pos_id=pos_id,
            report_name=report_name,
            **kwargs
        )

        reply = self.message_bus.send_message(
            component="nReports",
            message=Message(
                token=TK_NREPORTS_BUILD_REPORT,
                data_type=DataType.json,
                data=request,
            )
        )

        if reply.token == DefaultToken.TK_SYS_ACK:
            return reply.data

        raise self._build_exception_from_reply(reply=reply)

    def canceled_fiscal_coupon(
        self,           # type: NReports
        pos_id,         # type: int
        order_id,       # type: int
        is_mobile=False,   # type: Optional[bool]
    ):
        # type: (...) -> str

        return self.request_report(
            pos_id=pos_id,
            report_name=Reports.canceled_fiscal_coupon,
            order_id=order_id,
            is_mobile=is_mobile,
        )

    def canceled_receipt(
        self,               # type: NReports
        pos_id,             # type: int
        order_id,           # type: int
        is_mobile=False,    # type: Optional[bool]
    ):
        # type: (...) -> str

        return self.request_report(
            pos_id=pos_id,
            report_name=Reports.canceled_receipt,
            order_id=order_id,
            is_mobile=is_mobile,
        )

    def end_of_sale(
        self,           # type: NReports
        pos_id,         # type: int
        order_id,       # type: int
        mobile=False,   # type: Optional[bool]
    ):
        # type: (...) -> str

        return self.request_report(
            pos_id=pos_id,
            report_name=Reports.end_of_sale,
            order_id=order_id,
            is_mobile=mobile,
        )

    def pre_account(
        self,                                   # type: NReports
        pos_id,                                 # type: int
        order_id=None,                          # type: Optional[int]
        table_id=None,                          # type: Optional[int]
        is_mobile=False,                        # type: Optional[bool]
    ):
        # type: (...) -> str

        return self.request_report(
            pos_id=pos_id,
            report_name=Reports.pre_account,
            order_id=order_id,
            table_id=table_id,
            is_mobile=is_mobile,
        )

    def pre_account_quick_service(
        self,               # type: NReports
        pos_id,             # type: int
        order_id,           # type: int
        is_mobile=False,    # type: Optional[bool]
    ):
        # type: (...) -> str

        return self.request_report(
            pos_id=pos_id,
            report_name=Reports.pre_account_quick_service,
            order_id=order_id,
            is_mobile=is_mobile,
        )

    def pre_account_by_seats(
        self,                 # type: NReports
        pos_id,               # type: int
        table_id,             # type: str
        filter_table_seat,    # type: str
        print_by,             # type: str
        is_mobile,            # type: bool
    ):
        # type: (...) -> str

        return self.request_report(
            pos_id=pos_id,
            report_name=Reports.pre_account_by_seats,
            table_id=table_id,
            print_by=print_by,
            filter_table_seat=filter_table_seat,
            is_mobile=is_mobile,
        )

    def conference_items_by_seats(
        self,                 # type: NReports
        pos_id,               # type: int
        table_id,             # type: str
        filter_table_seat,    # type: str
        print_by,             # type: str
        is_mobile,            # type: bool
    ):
        # type: (...) -> str

        return self.request_report(
            pos_id=pos_id,
            report_name=Reports.conference_items_by_seats,
            table_id=table_id,
            print_by=print_by,
            filter_table_seat=filter_table_seat,
            is_mobile=is_mobile,
        )

    def conference_items(
        self,                                   # type: NReports
        pos_id,                                 # type: int
        order_id=None,                          # type: Optional[int]
        table_id=None,                          # type: Optional[int]
        is_mobile=False,                        # type: Optional[bool]
    ):
        # type: (...) -> str

        return self.request_report(
            pos_id=pos_id,
            report_name=Reports.conference_items,
            order_id=order_id,
            table_id=table_id,
            is_mobile=is_mobile,
        )

    def _build_request_data(
        self,           # type: NReports
        pos_id,         # type: int
        report_name,    # type: str
        **kwargs
    ):
        # type: (...) -> bytes

        request = {
            "posId": int(pos_id),
            "reportName": report_name,
        }
        for key, value in kwargs.items():
            request[self._to_camel_case(key)] = value

        return json.dumps(request).encode("utf-8")

    @staticmethod
    def _build_exception_from_reply(
        reply,  # type: Message
    ):
        # type: (...) -> NReportsException

        try:
            data_text = reply.data.decode("utf-8")
        except UnicodeDecodeError:
            data_text = reply.data.hex()

        try:
            response_dict = json.loads(data_text)
            return NReportsException(
                message=response_dict["errorMessage"],
                error_code=response_dict["errorCode"],
            )
        except (Exception,):
            return NReportsException(
                message=data_text,
                error_code=error.UNEXPECTED_ERROR,
            )

    @staticmethod
    def _to_camel_case(
        snake_str,  # type: str
    ):
        # type: (...) -> str

        components = snake_str.split("_")
        return components[0].lower() + "".join(x.lower().title() for x in components[1:])



import json

from ._Message import Message
from ._DefaultToken import DefaultToken
from ._DataType import DataType


class ApplicationErrorMessage(Message):
    def __init__(
        self,               # type: ApplicationErrorMessage
        error_code,         # type: int
        error_message,      # type: str
    ):
        self.error_code = error_code
        self.error_message = error_message

        super(ApplicationErrorMessage, self).__init__(
            token=DefaultToken.TK_SYS_APPLICATION_ERROR,
            data_type=DataType.json,
            data=json.dumps({
                "errorCode": error_code,
                "errorMessage": error_message,
            }).encode("utf-8"),
        )

    @staticmethod
    def build_from_message(
        message,        # type: Message
    ):
        if message.token != DefaultToken.TK_SYS_APPLICATION_ERROR:
            raise RuntimeError("Can only convert TK_SYS_APPLICATION_ERROR messages")

        bad_request_dict = json.loads(message.data.decode("utf-8"))
        return ApplicationErrorMessage(
            error_code=bad_request_dict["errorCode"],
            error_message=bad_request_dict["errorMessage"],
        )



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging


HISTORY_RECORDER = None
logger = logging.getLogger(__name__)


class BaseHistoryHandler(object):
    def emit(self, event_type, payload, source):
        raise NotImplementedError('emit()')


class HistoryRecorder(object):
    def __init__(self):
        self._enabled = False
        self._handlers = []

    def enable(self):
        self._enabled = True

    def disable(self):
        self._enabled = False

    def add_handler(self, handler):
        self._handlers.append(handler)

    def record(self, event_type, payload, source='BOTOCORE'):
        if self._enabled and self._handlers:
            for handler in self._handlers:
                try:
                    handler.emit(event_type, payload, source)
                except Exception:
                    # Never let the process die because we had a failure in
                    # a record collection handler.
                    logger.debug("Exception raised in %s.", handler,
                                 exc_info=True)


def get_global_history_recorder():
    global HISTORY_RECORDER
    if HISTORY_RECORDER is None:
        HISTORY_RECORDER = HistoryRecorder()
    return HISTORY_RECORDER



from helper import ExtendedEnum
from messagebus import (
    TokenCreator,
    TokenPriority,
)

DEVICE_MANAGER_GROUP = "D"


class ListenedTokens(ExtendedEnum):
    # Processors
    TK_GET_USED_SERVICE = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=DEVICE_MANAGER_GROUP,
        number="1",
    )
    TK_GET_DEVICES = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=DEVICE_MANAGER_GROUP,
        number="2",
    )
    TK_UPDATE_DEVICE_STATUS = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=DEVICE_MANAGER_GROUP,
        number="3",
    )



from msgbus import MBEasyContext
from old_helper import BaseRepository
from typing import List
from persistence import Connection


class ProductRepository(BaseRepository):
    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None
        super(ProductRepository, self).__init__(mb_context)

    def get_products_with_not_charge_tip(self):
        # type: () -> List[int]
        def inner_func(conn):
            # type: (Connection) -> List[int]
            return [int(x.get_entry(0)) for x in conn.select(self._GetProductsWithNotChargeTip)]

        return self.execute_with_connection(inner_func)

    _GetProductsWithNotChargeTip = \
        """SELECT ProductCode
           FROM ProductCustomParams
           WHERE CustomParamId = 'NotChargeTip' AND CustomParamValue = 'True'
        """



######################## BEGIN LICENSE BLOCK ########################
# The Original Code is mozilla.org code.
#
# The Initial Developer of the Original Code is
# Netscape Communications Corporation.
# Portions created by the Initial Developer are Copyright (C) 1998
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Mark Pilgrim - port to Python
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA
######################### END LICENSE BLOCK #########################

import logging

from .enums import MachineState


class CodingStateMachine(object):
    """
    A state machine to verify a byte sequence for a particular encoding. For
    each byte the detector receives, it will feed that byte to every active
    state machine available, one byte at a time. The state machine changes its
    state based on its previous state and the byte it receives. There are 3
    states in a state machine that are of interest to an auto-detector:

    START state: This is the state to start with, or a legal byte sequence
                 (i.e. a valid code point) for character has been identified.

    ME state:  This indicates that the state machine identified a byte sequence
               that is specific to the charset it is designed for and that
               there is no other possible encoding which can contain this byte
               sequence. This will to lead to an immediate positive answer for
               the detector.

    ERROR state: This indicates the state machine identified an illegal byte
                 sequence for that encoding. This will lead to an immediate
                 negative answer for this encoding. Detector will exclude this
                 encoding from consideration from here on.
    """
    def __init__(self, sm):
        self._model = sm
        self._curr_byte_pos = 0
        self._curr_char_len = 0
        self._curr_state = None
        self.logger = logging.getLogger(__name__)
        self.reset()

    def reset(self):
        self._curr_state = MachineState.START

    def next_state(self, c):
        # for each byte we get its class
        # if it is first byte, we also get byte length
        byte_class = self._model['class_table'][c]
        if self._curr_state == MachineState.START:
            self._curr_byte_pos = 0
            self._curr_char_len = self._model['char_len_table'][byte_class]
        # from byte's class and state_table, we get its next state
        curr_state = (self._curr_state * self._model['class_factor']
                      + byte_class)
        self._curr_state = self._model['state_table'][curr_state]
        self._curr_byte_pos += 1
        return self._curr_state

    def get_current_charlen(self):
        return self._curr_char_len

    def get_coding_state_machine(self):
        return self._model['name']

    @property
    def language(self):
        return self._model['language']



from ._LoyaltyTranslatedMessage import LoyaltyTranslatedMessage


class LoyaltyTotaledMessage(object):
    def __init__(
        self,                        # type: LoyaltyTotaledMessage
        loyalty_translated_message,  # type: LoyaltyTranslatedMessage
    ):
        # type: (...) -> None

        self.totalization_message = loyalty_translated_message



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore

"""
OpenTelemetry Base Instrumentor
"""

from abc import ABC, abstractmethod
from logging import getLogger
from typing import Collection, Optional

from opentelemetry.instrumentation.dependencies import (
    DependencyConflict,
    get_dependency_conflicts,
)

_LOG = getLogger(__name__)


class BaseInstrumentor(ABC):
    """An ABC for instrumentors

    Child classes of this ABC should instrument specific third
    party libraries or frameworks either by using the
    ``opentelemetry-instrument`` command or by calling their methods
    directly.

    Since every third party library or framework is different and has different
    instrumentation needs, more methods can be added to the child classes as
    needed to provide practical instrumentation to the end user.
    """

    _instance = None
    _is_instrumented_by_opentelemetry = False

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = object.__new__(cls)

        return cls._instance

    @property
    def is_instrumented_by_opentelemetry(self):
        return self._is_instrumented_by_opentelemetry

    @abstractmethod
    def instrumentation_dependencies(self) -> Collection[str]:
        """Return a list of python packages with versions that the will be instrumented.

        The format should be the same as used in requirements.txt or pyproject.toml.

        For example, if an instrumentation instruments requests 1.x, this method should look
        like:

            def instrumentation_dependencies(self) -> Collection[str]:
                return ['requests ~= 1.0']

        This will ensure that the instrumentation will only be used when the specified library
        is present in the environment.
        """

    def _instrument(self, **kwargs):
        """Instrument the library"""

    @abstractmethod
    def _uninstrument(self, **kwargs):
        """Uninstrument the library"""

    def _check_dependency_conflicts(self) -> Optional[DependencyConflict]:
        dependencies = self.instrumentation_dependencies()
        return get_dependency_conflicts(dependencies)

    def instrument(self, **kwargs):
        """Instrument the library

        This method will be called without any optional arguments by the
        ``opentelemetry-instrument`` command.

        This means that calling this method directly without passing any
        optional values should do the very same thing that the
        ``opentelemetry-instrument`` command does.
        """

        if self._is_instrumented_by_opentelemetry:
            _LOG.warning("Attempting to instrument while already instrumented")
            return None

        # check if instrumentor has any missing or conflicting dependencies
        skip_dep_check = kwargs.pop("skip_dep_check", False)
        if not skip_dep_check:
            conflict = self._check_dependency_conflicts()
            if conflict:
                _LOG.error(conflict)
                return None

        result = self._instrument(  # pylint: disable=assignment-from-no-return
            **kwargs
        )
        self._is_instrumented_by_opentelemetry = True
        return result

    def uninstrument(self, **kwargs):
        """Uninstrument the library

        See ``BaseInstrumentor.instrument`` for more information regarding the
        usage of ``kwargs``.
        """

        if self._is_instrumented_by_opentelemetry:
            result = self._uninstrument(**kwargs)
            self._is_instrumented_by_opentelemetry = False
            return result

        _LOG.warning("Attempting to uninstrument while already uninstrumented")

        return None


__all__ = ["BaseInstrumentor"]



from mw_helper import ensure_iterable

from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class OrderProductCategoryFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, allowed_category, forbidden_category, filter_method, logger=None):
        super(OrderProductCategoryFilterBox, self).__init__(name, sons, logger)
        self.allowed_category = ensure_iterable(allowed_category)
        self.forbidden_category = ensure_iterable(forbidden_category)
        self.filter_method = filter_method

    def change_order(self, order):
        new_order_items = []
        change = False

        allow, forbid = self.any_item_filter(order)

        if self.filter_method == "any":
            change = True

            if any(allow):
                new_order_items = order.items
            elif any(forbid):
                new_order_items = []

        if self.filter_method == "all":
            if any(allow):
                change = True
                new_order_items = order.items

            if any(forbid):
                change = True
                new_order_items = []

        if self.filter_method == "only":
            if any(allow):
                change = True
                new_order_items = [only for only in allow if only in order.items]

            if any(forbid):
                change = True
                new_order_items = filter(
                    lambda x: x not in [other for other in forbid if other in order.items],
                    order.items
                )

        if change:
            order.items = new_order_items

        return order

    def any_item_filter(self, order):
        allow_items = []
        forbid_items = []

        def alter(items, category_list, alter_list):
            for item in items:
                if item.items:
                    alter(item.items, category_list, alter_list)

                for category in category_list:
                    if category and category in item.json_tags:
                        alter_list.append(item)

        alter(order.items, self.allowed_category, allow_items)
        alter(order.items, self.forbidden_category, forbid_items)

        return allow_items, forbid_items



# -*- coding: utf-8 -*-
"""
Classic deprecation warning
===========================

Classic ``@deprecated`` decorator to deprecate old python classes, functions or methods.

.. _The Warnings Filter: https://docs.python.org/3/library/warnings.html#the-warnings-filter
"""
import functools
import inspect
import platform
import warnings

import wrapt

try:
    # If the C extension for wrapt was compiled and wrapt/_wrappers.pyd exists, then the
    # stack level that should be passed to warnings.warn should be 2. However, if using
    # a pure python wrapt, a extra stacklevel is required.
    import wrapt._wrappers

    _routine_stacklevel = 2
    _class_stacklevel = 2
except ImportError:
    _routine_stacklevel = 3
    if platform.python_implementation() == "PyPy":
        _class_stacklevel = 2
    else:
        _class_stacklevel = 3

string_types = (type(b''), type(u''))


class ClassicAdapter(wrapt.AdapterFactory):
    """
    Classic adapter -- *for advanced usage only*

    This adapter is used to get the deprecation message according to the wrapped object type:
    class, function, standard method, static method, or class method.

    This is the base class of the :class:`~deprecated.sphinx.SphinxAdapter` class
    which is used to update the wrapped object docstring.

    You can also inherit this class to change the deprecation message.

    In the following example, we change the message into "The ... is deprecated.":

    .. code-block:: python

       import inspect

       from deprecated.classic import ClassicAdapter
       from deprecated.classic import deprecated


       class MyClassicAdapter(ClassicAdapter):
           def get_deprecated_msg(self, wrapped, instance):
               if instance is None:
                   if inspect.isclass(wrapped):
                       fmt = "The class {name} is deprecated."
                   else:
                       fmt = "The function {name} is deprecated."
               else:
                   if inspect.isclass(instance):
                       fmt = "The class method {name} is deprecated."
                   else:
                       fmt = "The method {name} is deprecated."
               if self.reason:
                   fmt += " ({reason})"
               if self.version:
                   fmt += " -- Deprecated since version {version}."
               return fmt.format(name=wrapped.__name__,
                                 reason=self.reason or "",
                                 version=self.version or "")

    Then, you can use your ``MyClassicAdapter`` class like this in your source code:

    .. code-block:: python

       @deprecated(reason="use another function", adapter_cls=MyClassicAdapter)
       def some_old_function(x, y):
           return x + y
    """

    def __init__(self, reason="", version="", action=None, category=DeprecationWarning):
        """
        Construct a wrapper adapter.

        :type  reason: str
        :param reason:
            Reason message which documents the deprecation in your library (can be omitted).

        :type  version: str
        :param version:
            Version of your project which deprecates this feature.
            If you follow the `Semantic Versioning <https://semver.org/>`_,
            the version number has the format "MAJOR.MINOR.PATCH".

        :type  action: str
        :param action:
            A warning filter used to activate or not the deprecation warning.
            Can be one of "error", "ignore", "always", "default", "module", or "once".
            If ``None`` or empty, the the global filtering mechanism is used.
            See: `The Warnings Filter`_ in the Python documentation.

        :type  category: type
        :param category:
            The warning category to use for the deprecation warning.
            By default, the category class is :class:`~DeprecationWarning`,
            you can inherit this class to define your own deprecation warning category.
        """
        self.reason = reason or ""
        self.version = version or ""
        self.action = action
        self.category = category
        super(ClassicAdapter, self).__init__()

    def get_deprecated_msg(self, wrapped, instance):
        """
        Get the deprecation warning message for the user.

        :param wrapped: Wrapped class or function.

        :param instance: The object to which the wrapped function was bound when it was called.

        :return: The warning message.
        """
        if instance is None:
            if inspect.isclass(wrapped):
                fmt = "Call to deprecated class {name}."
            else:
                fmt = "Call to deprecated function (or staticmethod) {name}."
        else:
            if inspect.isclass(instance):
                fmt = "Call to deprecated class method {name}."
            else:
                fmt = "Call to deprecated method {name}."
        if self.reason:
            fmt += " ({reason})"
        if self.version:
            fmt += " -- Deprecated since version {version}."
        return fmt.format(name=wrapped.__name__, reason=self.reason or "", version=self.version or "")

    def __call__(self, wrapped):
        """
        Decorate your class or function.

        :param wrapped: Wrapped class or function.

        :return: the decorated class or function.

        .. versionchanged:: 1.2.4
           Don't pass arguments to :meth:`object.__new__` (other than *cls*).

        .. versionchanged:: 1.2.8
           The warning filter is not set if the *action* parameter is ``None`` or empty.
        """
        if inspect.isclass(wrapped):
            old_new1 = wrapped.__new__

            def wrapped_cls(cls, *args, **kwargs):
                msg = self.get_deprecated_msg(wrapped, None)
                if self.action:
                    with warnings.catch_warnings():
                        warnings.simplefilter(self.action, self.category)
                        warnings.warn(msg, category=self.category, stacklevel=_class_stacklevel)
                else:
                    warnings.warn(msg, category=self.category, stacklevel=_class_stacklevel)
                if old_new1 is object.__new__:
                    return old_new1(cls)
                # actually, we don't know the real signature of *old_new1*
                return old_new1(cls, *args, **kwargs)

            wrapped.__new__ = staticmethod(wrapped_cls)

        return wrapped


def deprecated(*args, **kwargs):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.

    **Classic usage:**

    To use this, decorate your deprecated function with **@deprecated** decorator:

    .. code-block:: python

       from deprecated import deprecated


       @deprecated
       def some_old_function(x, y):
           return x + y

    You can also decorate a class or a method:

    .. code-block:: python

       from deprecated import deprecated


       class SomeClass(object):
           @deprecated
           def some_old_method(self, x, y):
               return x + y


       @deprecated
       class SomeOldClass(object):
           pass

    You can give a *reason* message to help the developer to choose another function/class,
    and a *version* number to specify the starting version number of the deprecation.

    .. code-block:: python

       from deprecated import deprecated


       @deprecated(reason="use another function", version='1.2.0')
       def some_old_function(x, y):
           return x + y

    The *category* keyword argument allow you to specify the deprecation warning class of your choice.
    By default, :exc:`DeprecationWarning` is used but you can choose :exc:`FutureWarning`,
    :exc:`PendingDeprecationWarning` or a custom subclass.

    .. code-block:: python

       from deprecated import deprecated


       @deprecated(category=PendingDeprecationWarning)
       def some_old_function(x, y):
           return x + y

    The *action* keyword argument allow you to locally change the warning filtering.
    *action* can be one of "error", "ignore", "always", "default", "module", or "once".
    If ``None``, empty or missing, the the global filtering mechanism is used.
    See: `The Warnings Filter`_ in the Python documentation.

    .. code-block:: python

       from deprecated import deprecated


       @deprecated(action="error")
       def some_old_function(x, y):
           return x + y

    """
    if args and isinstance(args[0], string_types):
        kwargs['reason'] = args[0]
        args = args[1:]

    if args and not callable(args[0]):
        raise TypeError(repr(type(args[0])))

    if args:
        action = kwargs.get('action')
        category = kwargs.get('category', DeprecationWarning)
        adapter_cls = kwargs.pop('adapter_cls', ClassicAdapter)
        adapter = adapter_cls(**kwargs)

        wrapped = args[0]
        if inspect.isclass(wrapped):
            wrapped = adapter(wrapped)
            return wrapped

        elif inspect.isroutine(wrapped):

            @wrapt.decorator(adapter=adapter)
            def wrapper_function(wrapped_, instance_, args_, kwargs_):
                msg = adapter.get_deprecated_msg(wrapped_, instance_)
                if action:
                    with warnings.catch_warnings():
                        warnings.simplefilter(action, category)
                        warnings.warn(msg, category=category, stacklevel=_routine_stacklevel)
                else:
                    warnings.warn(msg, category=category, stacklevel=_routine_stacklevel)
                return wrapped_(*args_, **kwargs_)

            return wrapper_function(wrapped)

        else:
            raise TypeError(repr(type(wrapped)))

    return functools.partial(deprecated, **kwargs)



# nric.py - functions for handling  NRIC numbers
#
# Copyright (C) 2013, 2014 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NRIC No. (Malaysian National Registration Identity Card Number).

The NRIC No. is the unique identifier for issued to Malaysian citizens and
permanent residents.

The number consist of 12 digits in three sections. The first 6 digits
represent the birth date, followed by two digits representing the birth
place and finally four digits. The gender of a person can be derived from
the last digit: odd numbers for males and even numbers for females.

>>> validate('770305-02-1234')
'770305021234'
>>> validate('771305-02-1234')  # invalid date
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('770305-17-1234')  # unknown birth place code
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> format('770305021234')
'770305-02-1234'
"""

import datetime

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -*').strip()


def get_birth_date(number):
    """Split the date parts from the number and return the birth date.
    Note that in some cases it may return the registration date instead of
    the birth date and it may be a century off."""
    number = compact(number)
    year = int(number[0:2])
    month = int(number[2:4])
    day = int(number[4:6])
    # this is a bit broken but it's easy
    try:
        return datetime.date(year + 1900, month, day)
    except ValueError:
        pass
    try:
        return datetime.date(year + 2000, month, day)
    except ValueError:
        raise InvalidComponent()


def get_birth_place(number):
    """Use the number to look up the place of birth of the person. This can
    either be a state or federal territory within Malaysia or a country
    outside of Malaysia."""
    from stdnum import numdb
    number = compact(number)
    results = numdb.get('my/bp').info(number[6:8])[0][1]
    if not results:
        raise InvalidComponent()
    return results


def validate(number):
    """Check if the number is a valid NRIC number. This checks the length,
    formatting and birth date and place."""
    number = compact(number)
    if len(number) != 12:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    get_birth_date(number)
    get_birth_place(number)
    return number


def is_valid(number):
    """Check if the number is a valid NRIC number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return number[:6] + '-' + number[6:8] + '-' + number[8:]



from ._SequencerService import SequencerService
from ._OrderService import OrderService
from ._FeeService import (
    FeeService,
    get_order_fees,
    get_fee_details,
    order_xml_has_valid_fee,
)
from ._FiscalOrderValidationService import FiscalOrderValidationService



from ._OrderChangerProductionBox import OrderChangerProductionBox


class TagItemsWhenProdStateBox(OrderChangerProductionBox):
    def __init__(self, name, sons, prod_state, tag, logger=None):
        super(TagItemsWhenProdStateBox, self).__init__(name, sons, logger)
        self.prod_state = prod_state
        self.tag = tag

    def change_order(self, order):
        if order.prod_state == self.prod_state:
            self._tag_items(order.items)

        return order

    def _tag_items(self, items):
        for item in items:
            item.add_tag(self.tag)
            
            if item.items:
                self._tag_items(item.items)



from ._UndoCommand import UndoCommand
from ._TagLinesUndoCommand import TagLinesUndoCommand
from ._ChangeProdStateUndoCommand import ChangeProdStateUndoCommand
from ._TagLinesUndoCommandType import TagLinesUndoCommandType



######################## BEGIN LICENSE BLOCK ########################
# The Original Code is mozilla.org code.
#
# The Initial Developer of the Original Code is
# Netscape Communications Corporation.
# Portions created by the Initial Developer are Copyright (C) 1998
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Mark Pilgrim - port to Python
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA
######################### END LICENSE BLOCK #########################

from .enums import ProbingState, MachineState
from .mbcharsetprober import MultiByteCharSetProber
from .codingstatemachine import CodingStateMachine
from .chardistribution import EUCJPDistributionAnalysis
from .jpcntx import EUCJPContextAnalysis
from .mbcssm import EUCJP_SM_MODEL


class EUCJPProber(MultiByteCharSetProber):
    def __init__(self):
        super(EUCJPProber, self).__init__()
        self.coding_sm = CodingStateMachine(EUCJP_SM_MODEL)
        self.distribution_analyzer = EUCJPDistributionAnalysis()
        self.context_analyzer = EUCJPContextAnalysis()
        self.reset()

    def reset(self):
        super(EUCJPProber, self).reset()
        self.context_analyzer.reset()

    @property
    def charset_name(self):
        return "EUC-JP"

    @property
    def language(self):
        return "Japanese"

    def feed(self, byte_str):
        for i in range(len(byte_str)):
            # PY3K: byte_str is a byte array, so byte_str[i] is an int, not a byte
            coding_state = self.coding_sm.next_state(byte_str[i])
            if coding_state == MachineState.ERROR:
                self.logger.debug('%s %s prober hit error at byte %s',
                                  self.charset_name, self.language, i)
                self._state = ProbingState.NOT_ME
                break
            elif coding_state == MachineState.ITS_ME:
                self._state = ProbingState.FOUND_IT
                break
            elif coding_state == MachineState.START:
                char_len = self.coding_sm.get_current_charlen()
                if i == 0:
                    self._last_char[1] = byte_str[0]
                    self.context_analyzer.feed(self._last_char, char_len)
                    self.distribution_analyzer.feed(self._last_char, char_len)
                else:
                    self.context_analyzer.feed(byte_str[i - 1:i + 1],
                                                char_len)
                    self.distribution_analyzer.feed(byte_str[i - 1:i + 1],
                                                     char_len)

        self._last_char[0] = byte_str[-1]

        if self.state == ProbingState.DETECTING:
            if (self.context_analyzer.got_enough_data() and
               (self.get_confidence() > self.SHORTCUT_THRESHOLD)):
                self._state = ProbingState.FOUND_IT

        return self.state

    def get_confidence(self):
        context_conf = self.context_analyzer.get_confidence()
        distrib_conf = self.distribution_analyzer.get_confidence()
        return max(context_conf, distrib_conf)



data = (
'ssal',    # 0x00
'ssalg',    # 0x01
'ssalm',    # 0x02
'ssalb',    # 0x03
'ssals',    # 0x04
'ssalt',    # 0x05
'ssalp',    # 0x06
'ssalh',    # 0x07
'ssam',    # 0x08
'ssab',    # 0x09
'ssabs',    # 0x0a
'ssas',    # 0x0b
'ssass',    # 0x0c
'ssang',    # 0x0d
'ssaj',    # 0x0e
'ssac',    # 0x0f
'ssak',    # 0x10
'ssat',    # 0x11
'ssap',    # 0x12
'ssah',    # 0x13
'ssae',    # 0x14
'ssaeg',    # 0x15
'ssaegg',    # 0x16
'ssaegs',    # 0x17
'ssaen',    # 0x18
'ssaenj',    # 0x19
'ssaenh',    # 0x1a
'ssaed',    # 0x1b
'ssael',    # 0x1c
'ssaelg',    # 0x1d
'ssaelm',    # 0x1e
'ssaelb',    # 0x1f
'ssaels',    # 0x20
'ssaelt',    # 0x21
'ssaelp',    # 0x22
'ssaelh',    # 0x23
'ssaem',    # 0x24
'ssaeb',    # 0x25
'ssaebs',    # 0x26
'ssaes',    # 0x27
'ssaess',    # 0x28
'ssaeng',    # 0x29
'ssaej',    # 0x2a
'ssaec',    # 0x2b
'ssaek',    # 0x2c
'ssaet',    # 0x2d
'ssaep',    # 0x2e
'ssaeh',    # 0x2f
'ssya',    # 0x30
'ssyag',    # 0x31
'ssyagg',    # 0x32
'ssyags',    # 0x33
'ssyan',    # 0x34
'ssyanj',    # 0x35
'ssyanh',    # 0x36
'ssyad',    # 0x37
'ssyal',    # 0x38
'ssyalg',    # 0x39
'ssyalm',    # 0x3a
'ssyalb',    # 0x3b
'ssyals',    # 0x3c
'ssyalt',    # 0x3d
'ssyalp',    # 0x3e
'ssyalh',    # 0x3f
'ssyam',    # 0x40
'ssyab',    # 0x41
'ssyabs',    # 0x42
'ssyas',    # 0x43
'ssyass',    # 0x44
'ssyang',    # 0x45
'ssyaj',    # 0x46
'ssyac',    # 0x47
'ssyak',    # 0x48
'ssyat',    # 0x49
'ssyap',    # 0x4a
'ssyah',    # 0x4b
'ssyae',    # 0x4c
'ssyaeg',    # 0x4d
'ssyaegg',    # 0x4e
'ssyaegs',    # 0x4f
'ssyaen',    # 0x50
'ssyaenj',    # 0x51
'ssyaenh',    # 0x52
'ssyaed',    # 0x53
'ssyael',    # 0x54
'ssyaelg',    # 0x55
'ssyaelm',    # 0x56
'ssyaelb',    # 0x57
'ssyaels',    # 0x58
'ssyaelt',    # 0x59
'ssyaelp',    # 0x5a
'ssyaelh',    # 0x5b
'ssyaem',    # 0x5c
'ssyaeb',    # 0x5d
'ssyaebs',    # 0x5e
'ssyaes',    # 0x5f
'ssyaess',    # 0x60
'ssyaeng',    # 0x61
'ssyaej',    # 0x62
'ssyaec',    # 0x63
'ssyaek',    # 0x64
'ssyaet',    # 0x65
'ssyaep',    # 0x66
'ssyaeh',    # 0x67
'sseo',    # 0x68
'sseog',    # 0x69
'sseogg',    # 0x6a
'sseogs',    # 0x6b
'sseon',    # 0x6c
'sseonj',    # 0x6d
'sseonh',    # 0x6e
'sseod',    # 0x6f
'sseol',    # 0x70
'sseolg',    # 0x71
'sseolm',    # 0x72
'sseolb',    # 0x73
'sseols',    # 0x74
'sseolt',    # 0x75
'sseolp',    # 0x76
'sseolh',    # 0x77
'sseom',    # 0x78
'sseob',    # 0x79
'sseobs',    # 0x7a
'sseos',    # 0x7b
'sseoss',    # 0x7c
'sseong',    # 0x7d
'sseoj',    # 0x7e
'sseoc',    # 0x7f
'sseok',    # 0x80
'sseot',    # 0x81
'sseop',    # 0x82
'sseoh',    # 0x83
'sse',    # 0x84
'sseg',    # 0x85
'ssegg',    # 0x86
'ssegs',    # 0x87
'ssen',    # 0x88
'ssenj',    # 0x89
'ssenh',    # 0x8a
'ssed',    # 0x8b
'ssel',    # 0x8c
'sselg',    # 0x8d
'sselm',    # 0x8e
'sselb',    # 0x8f
'ssels',    # 0x90
'sselt',    # 0x91
'sselp',    # 0x92
'sselh',    # 0x93
'ssem',    # 0x94
'sseb',    # 0x95
'ssebs',    # 0x96
'sses',    # 0x97
'ssess',    # 0x98
'sseng',    # 0x99
'ssej',    # 0x9a
'ssec',    # 0x9b
'ssek',    # 0x9c
'sset',    # 0x9d
'ssep',    # 0x9e
'sseh',    # 0x9f
'ssyeo',    # 0xa0
'ssyeog',    # 0xa1
'ssyeogg',    # 0xa2
'ssyeogs',    # 0xa3
'ssyeon',    # 0xa4
'ssyeonj',    # 0xa5
'ssyeonh',    # 0xa6
'ssyeod',    # 0xa7
'ssyeol',    # 0xa8
'ssyeolg',    # 0xa9
'ssyeolm',    # 0xaa
'ssyeolb',    # 0xab
'ssyeols',    # 0xac
'ssyeolt',    # 0xad
'ssyeolp',    # 0xae
'ssyeolh',    # 0xaf
'ssyeom',    # 0xb0
'ssyeob',    # 0xb1
'ssyeobs',    # 0xb2
'ssyeos',    # 0xb3
'ssyeoss',    # 0xb4
'ssyeong',    # 0xb5
'ssyeoj',    # 0xb6
'ssyeoc',    # 0xb7
'ssyeok',    # 0xb8
'ssyeot',    # 0xb9
'ssyeop',    # 0xba
'ssyeoh',    # 0xbb
'ssye',    # 0xbc
'ssyeg',    # 0xbd
'ssyegg',    # 0xbe
'ssyegs',    # 0xbf
'ssyen',    # 0xc0
'ssyenj',    # 0xc1
'ssyenh',    # 0xc2
'ssyed',    # 0xc3
'ssyel',    # 0xc4
'ssyelg',    # 0xc5
'ssyelm',    # 0xc6
'ssyelb',    # 0xc7
'ssyels',    # 0xc8
'ssyelt',    # 0xc9
'ssyelp',    # 0xca
'ssyelh',    # 0xcb
'ssyem',    # 0xcc
'ssyeb',    # 0xcd
'ssyebs',    # 0xce
'ssyes',    # 0xcf
'ssyess',    # 0xd0
'ssyeng',    # 0xd1
'ssyej',    # 0xd2
'ssyec',    # 0xd3
'ssyek',    # 0xd4
'ssyet',    # 0xd5
'ssyep',    # 0xd6
'ssyeh',    # 0xd7
'sso',    # 0xd8
'ssog',    # 0xd9
'ssogg',    # 0xda
'ssogs',    # 0xdb
'sson',    # 0xdc
'ssonj',    # 0xdd
'ssonh',    # 0xde
'ssod',    # 0xdf
'ssol',    # 0xe0
'ssolg',    # 0xe1
'ssolm',    # 0xe2
'ssolb',    # 0xe3
'ssols',    # 0xe4
'ssolt',    # 0xe5
'ssolp',    # 0xe6
'ssolh',    # 0xe7
'ssom',    # 0xe8
'ssob',    # 0xe9
'ssobs',    # 0xea
'ssos',    # 0xeb
'ssoss',    # 0xec
'ssong',    # 0xed
'ssoj',    # 0xee
'ssoc',    # 0xef
'ssok',    # 0xf0
'ssot',    # 0xf1
'ssop',    # 0xf2
'ssoh',    # 0xf3
'sswa',    # 0xf4
'sswag',    # 0xf5
'sswagg',    # 0xf6
'sswags',    # 0xf7
'sswan',    # 0xf8
'sswanj',    # 0xf9
'sswanh',    # 0xfa
'sswad',    # 0xfb
'sswal',    # 0xfc
'sswalg',    # 0xfd
'sswalm',    # 0xfe
'sswalb',    # 0xff
)



from typing import Dict

from ._Message import Message
from ._DefaultToken import DefaultToken


class Event(object):
    def __init__(
        self,                   # type: Event
        subject,                # type: str
        event_type=None,        # type: str
        data=None,              # type: bytes
        source=0,               # type: int
        synchronous=False,      # type: bool
        timeout=-1,             # type: int
        headers=None,           # type: Dict[str, str]
        ttl=-1,                 # type: int
    ):
        # type: (...) -> None
        self.subject = subject
        self.event_type = event_type
        self.source = source
        self.data = data
        self.synchronous = synchronous
        self.timeout = timeout
        self.headers = headers if headers is not None else {}
        self.ttl = ttl

        self.imp_message = None

    def __str__(self):
        # type: () -> str
        return str({
            "subject": self.subject,
            "event_type": self.event_type,
            "synchronous": self.synchronous,
            "timeout": self.timeout,
            "headers": self.headers,
            "ttl": self.ttl,
        })

    @staticmethod
    def build_from_message(
        message,        # type: Message
    ):
        # type: (...) -> Event

        if message.token != DefaultToken.TK_EVT_EVENT:
            raise RuntimeError("Can only convert event messages")

        split_data = message.data.split(b"\0")
        event_data = split_data[0]
        subject = split_data[1].decode("utf-8")
        event_type = split_data[2].decode("utf-8")
        sync = split_data[3] == b"true"

        return Event(
            subject=subject,
            event_type=event_type,
            data=event_data,
            synchronous=sync,
            headers=message.headers,
        )

    def convert_to_message(
        self,       # type: Event
    ):
        # type: (...) -> Message

        data = b""
        if self.data is not None:
            data += self.data
        data += b"\0"
        data += self.subject.encode("utf-8") + b"\0"
        if self.event_type is not None:
            data += self.event_type.encode("utf-8")
        data += b"\0"
        data += b"true" if self.synchronous else b"false"

        return Message(
            token=DefaultToken.TK_EVT_EVENT,
            data=data,
            headers=self.headers,
        )



class Tender(object):

    def __init__(
        self,
        tender_id=None,     # type: int
        amount=None,        # type: float
    ):
        # type: (...) -> None

        self.tender_id = tender_id or -1
        self.amount = amount or 0.0



from _GetSalesChallengeInteractor import GetSalesChallengeInteractor
from _InsertSalesChallengeInteractor import InsertSalesChallengeInteractor
from _DeleteSalesChallengeInteractor import DeleteSalesChallengeInteractor
from _UpdateSalesChallengeInteractor import UpdateSalesChallengeInteractor



import logging
from logging import Logger

from cfgtools import read
from helper import config_logger
from model.config import Config


class Configurator(object):

    def __init__(self, loader_path):
        # type: (str) -> None

        self.config = Config()
        self.config.loader_path = loader_path
        self.reader = read(loader_path)

    def read_loader_configs(self):
        # type: () -> Config

        group = "AuditLogger."

        self.config.component_name = str(self.reader.find_value("Identification.ComponentName"))
        self.config.output_folder = self.reader.find_value(group + "output_folder")
        self.config.encode = (self.reader.find_value(group + "encode") or "true").lower() == "true"
        self.config.log_level = (self.reader.find_value(group + "LogLevel") or "INFO").upper()
        self.config.connection_timeout = int(self.reader.find_value(group + "ConnectionTimeout") or "5")

        return self.config

    def configure_logger(self):
        # type: () -> Logger

        config_logger(self.config.loader_path, self.config.component_name, root_level=logging.DEBUG)
        return logging.getLogger(self.config.component_name)



# defusedxml
#
# Copyright (c) 2013 by Christian Heimes <christian@python.org>
# Licensed to PSF under a Contributor Agreement.
# See https://www.python.org/psf/license for licensing details.
"""Defused xml.dom.pulldom
"""
from __future__ import print_function, absolute_import

from xml.dom.pulldom import parse as _parse
from xml.dom.pulldom import parseString as _parseString
from .sax import make_parser

__origin__ = "xml.dom.pulldom"


def parse(
    stream_or_string,
    parser=None,
    bufsize=None,
    forbid_dtd=False,
    forbid_entities=True,
    forbid_external=True,
):
    if parser is None:
        parser = make_parser()
        parser.forbid_dtd = forbid_dtd
        parser.forbid_entities = forbid_entities
        parser.forbid_external = forbid_external
    return _parse(stream_or_string, parser, bufsize)


def parseString(
    string, parser=None, forbid_dtd=False, forbid_entities=True, forbid_external=True
):
    if parser is None:
        parser = make_parser()
        parser.forbid_dtd = forbid_dtd
        parser.forbid_entities = forbid_entities
        parser.forbid_external = forbid_external
    return _parseString(string, parser)



from mw_helper import ensure_iterable

from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class FromOrderStateFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, states, logger=None):
        super(FromOrderStateFilterBox, self).__init__(name, sons, logger)
        self.states = ensure_iterable(states)

    def change_order(self, order):
        state_history = [x.state for x in order.state_history]
        for state in self.states:
            if state in state_history:
                return order

        return None



from abc import ABCMeta, abstractmethod
from datetime import datetime
from logging import Logger
from threading import Timer

from application.interactor import UserUpdaterInteractor, MediaUpdaterInteractor, LoaderUpdaterInteractor, \
    CatalogUpdaterInteractor, BinaryUpdaterInteractor
from application.model import BaseThread
from application.model.configuration import Configurations
from application.model.customexception import NotInsideUpdateWindow, NotImplementedException, NoUpdateFound, \
    DisabledFunctionality
from application.repository.pos import UserUpdaterPOSRepository, MediaUpdaterPOSRepository, \
    LoaderUpdaterPOSRepository, CatalogUpdaterPOSRepository, BinaryUpdaterPOSRepository
from application.util import ValidationUtil
from typing import Union, Optional


class UpdaterManager(BaseThread):
    __metaclass__ = ABCMeta

    def __init__(
        self,               # type: UpdaterManager
        update_name,        # type: str
        configs,            # type: Configurations
        interactor,         # type: Union[UserUpdaterInteractor, MediaUpdaterInteractor, LoaderUpdaterInteractor, CatalogUpdaterInteractor, BinaryUpdaterInteractor] # noqa
        pos_repository,     # type: Union[UserUpdaterPOSRepository, MediaUpdaterPOSRepository, LoaderUpdaterPOSRepository, CatalogUpdaterPOSRepository, BinaryUpdaterPOSRepository] # noqa
        logger,             # type: Logger
    ):
        # type: (...) -> None

        super(UpdaterManager, self).__init__()
        self.update_name = update_name
        self.local_configs = configs.updaters[self.update_name]
        self.active_frequency = int(self.local_configs.active_frequency)
        self.interactor = interactor
        self.pos_repository = pos_repository
        self.logger = logger

        self.update_scheduled = {}

    def run(
        self,                   # type: UpdaterManager
        validate_window=True,   # type: Optional[bool]
    ):
        # type: (...) -> None

        if self.active_frequency == 0:
            self.logger.info("{} - Auto-update thread not active. Stopping manager".format(self.update_name))
            self.stop()
            return

        self.logger.info("{} - update thread manager started".format(self.update_name))

        self._execute(validate_window=validate_window)

    def _execute(self, validate_window):
        # type: (bool) -> None

        disabled_update = False
        while self.running:
            try:
                self.notify_pending_update()
                self._has_scheduled_next_update()

                self._check_update_window(validate_window)
                self.has_update_to_apply()

                self.update()
            except DisabledFunctionality:
                self.logger.info("{} - Auto-update thread disabled. Stopping manager".format(self.update_name))
                disabled_update = True
                self.stop()
                return
            except NotInsideUpdateWindow:
                self.logger.info("{} - Not inside update windows".format(self.update_name))
            except NoUpdateFound:
                self.logger.info("{} - No update found".format(self.update_name))
            except (Exception,):
                self.logger.exception("{} - Not expected exception on update".format(self.update_name))
            finally:
                if not disabled_update:
                    if not self.pos_repository.has_already_applied_update():
                        self.sleep(30)
                    else:
                        self.sleep(self.active_frequency)

    @abstractmethod
    def update(self):
        # type: () -> None

        raise NotImplementedException()

    @abstractmethod
    def notify_pending_update(
        self,   # type: UpdaterManager
    ):
        # type: (...) -> None

        raise NotImplementedError()

    def has_update_to_apply(self):
        # type: () -> None

        return

    def get_pending_remote_update_to_schedule(
        self,                  # type: UpdaterManager
        external_update_id,    # type: str
    ):
        # type: (...) -> Optional[datetime.time]

        return

    def get_pending_remote_update_id(self):
        # type: (...) -> Optional[str]

        return

    def _check_update_window(self, validate_window):
        # type: (bool) -> None

        if not validate_window or not self.pos_repository.has_already_applied_update():
            return

        start_time = self.local_configs.window.start_time
        end_time = self.local_configs.window.end_time
        if not ValidationUtil().is_inside_allowed_window(start_time, end_time):
            str_start_time = start_time.strftime("%H:%M:%S")
            str_end_time = end_time.strftime("%H:%M:%S")
            message = "Not inside allowed window. Start: {} / End: {}".format(str_start_time, str_end_time)
            self.logger.info(message)
            raise NotInsideUpdateWindow()

    def _has_scheduled_next_update(
        self,   # type: UpdaterManager
    ):
        # type: (...) -> Optional[bool]

        external_update_id = self.get_pending_remote_update_id()
        update_time = self.get_pending_remote_update_to_schedule(external_update_id=external_update_id)
        if not update_time:
            return

        timer = Timer(
            interval=(update_time - datetime.now()).total_seconds(),
            function=self._execute_timer,
            args=(external_update_id, )
        )
        timer.start()

        self.logger.info("Remote Loader Update scheduled for {}".format(update_time))

    def _execute_timer(
        self,                  # type: UpdaterManager
        external_update_id,    # type: str
    ):
        # type: (...) -> None

        if external_update_id in self.update_scheduled:
            del self.update_scheduled[external_update_id]

        self.update()



# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/api/endpoint.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x19google/api/endpoint.proto\x12\ngoogle.api"Q\n\x08\x45ndpoint\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x07\x61liases\x18\x02 \x03(\tB\x02\x18\x01\x12\x0e\n\x06target\x18\x65 \x01(\t\x12\x12\n\nallow_cors\x18\x05 \x01(\x08\x42o\n\x0e\x63om.google.apiB\rEndpointProtoP\x01ZEgoogle.golang.org/genproto/googleapis/api/serviceconfig;serviceconfig\xa2\x02\x04GAPIb\x06proto3'
)


_ENDPOINT = DESCRIPTOR.message_types_by_name["Endpoint"]
Endpoint = _reflection.GeneratedProtocolMessageType(
    "Endpoint",
    (_message.Message,),
    {
        "DESCRIPTOR": _ENDPOINT,
        "__module__": "google.api.endpoint_pb2"
        # @@protoc_insertion_point(class_scope:google.api.Endpoint)
    },
)
_sym_db.RegisterMessage(Endpoint)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\016com.google.apiB\rEndpointProtoP\001ZEgoogle.golang.org/genproto/googleapis/api/serviceconfig;serviceconfig\242\002\004GAPI"
    _ENDPOINT.fields_by_name["aliases"]._options = None
    _ENDPOINT.fields_by_name["aliases"]._serialized_options = b"\030\001"
    _ENDPOINT._serialized_start = 41
    _ENDPOINT._serialized_end = 122
# @@protoc_insertion_point(module_scope)
